# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

import sys
import logging
import argparse
from decimal import Decimal

import dateutil.parser

from ..version import __version__
from ..config import config
from .valueasset import ValueAsset

if sys.version_info[0] >= 3:
    sys.stdout.reconfigure(encoding='utf-8')

logging.basicConfig(stream=sys.stdout,
                    level=logging.INFO,
                    format='[%(asctime)s.%(msecs)03d] %(levelname)s -- : %(message)s',
                    datefmt='%Y-%m-%dT%H:%M:%S')
log = logging.getLogger()

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("asset",
                        type=str,
                        nargs=1,
                        help="symbol of cryptoasset or fiat currency (i.e. BTC/LTC/ETH or EUR/USD)")
    parser.add_argument("date",
                        type=str,
                        nargs='?',
                        help="date (YYYY-MM-DD)")
    parser.add_argument("-v",
                        "--version",
                        action='version',
                        version='%(prog)s {version}'.format(version=__version__))
    parser.add_argument("-d",
                        "--debug",
                        action='store_true',
                        help="enabled debug logging")
    parser.add_argument("-q",
                        "--quantity",
                        type=Decimal,
                        help="quantity to price")
    parser.add_argument("-nc",
                        "--nocache",
                        action='store_true', help="bypass cache for historical data")

    config.args = parser.parse_args()

    if config.args.debug:
        log.setLevel(logging.DEBUG)
        logging.getLogger('urllib3').setLevel(logging.WARNING)
        config.output_config(parser.prog)

    value_asset = ValueAsset()
    asset = config.args.asset[0]
    timestamp = None

    if asset == config.CCY:
        return

    if config.args.date:
        timestamp = dateutil.parser.parse(config.args.date)
        timestamp = timestamp.replace(tzinfo=config.TZ_LOCAL)
        price_ccy, name, data_source = value_asset.get_historical_price(asset, timestamp)
    else:
        price_ccy, name, data_source = value_asset.get_latest_price(asset)

    if price_ccy is not None:
        log.info("1 %s=%s%s %s via %s (%s)",
                 asset,
                 config.sym(), '{:0,.2f}'.format(price_ccy), config.CCY,
                 data_source, name)
        if config.args.quantity:
            quantity = Decimal(config.args.quantity)
            log.info("%s %s=%s%s %s",
                     '{:0,f}'.format(quantity.normalize()),
                     asset,
                     config.sym(), '{:0,.2f}'.format(quantity * price_ccy), config.CCY)
    else:
        if name is not None:
            if timestamp:
                log.warning("Price for %s at %s is not available",
                            asset, timestamp.strftime('%Y-%m-%d'))
            else:
                log.warning("Current price for %s is not available", asset)
        else:
            log.warning("Prices for %s are not supported", asset)



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

import logging
import os
from datetime import datetime

import jinja2
from xhtml2pdf import pisa
import dateutil.parser

from .version import __version__
from .config import config

log = logging.getLogger()

class ReportPdf(object):
    DEFAULT_FILENAME = 'BittyTax_Report'
    FILE_EXTENSION = 'pdf'
    TEMPLATE_FILE = 'tax_report.html'

    def __init__(self, progname, audit, tax_report, price_report, holdings_report):
        self.env = jinja2.Environment(loader=jinja2.PackageLoader('bittytax', 'templates'))
        self.filename = self.get_output_filename(self.FILE_EXTENSION)

        self.env.filters['datefilter'] = self.datefilter
        self.env.filters['quantityfilter'] = self.quantityfilter
        self.env.filters['valuefilter'] = self.valuefilter
        self.env.filters['nowrapfilter'] = self.nowrapfilter

        template = self.env.get_template(self.TEMPLATE_FILE)
        html = template.render({'date': datetime.now(),
                                'author': '{} {}'.format(progname, __version__),
                                'config': config,
                                'audit': audit,
                                'tax_report': tax_report,
                                'price_report': price_report,
                                'holdings_report': holdings_report})

        log.info("Generating PDF report...")
        pdf_file = open(self.filename, 'w+b')
        status = pisa.CreatePDF(html, dest=pdf_file)
        pdf_file.close()

        if not status.err:
            log.info("PDF tax report file created: %s", self.filename)

    @staticmethod
    def datefilter(date):
        if isinstance(date, datetime):
            return date.strftime('%d/%m/%Y')
        else:
            return dateutil.parser.parse(date).strftime('%d/%m/%Y')

    @staticmethod
    def quantityfilter(quantity):
        return '{:0,f}'.format(quantity.normalize())

    @staticmethod
    def valuefilter(value):
        return '&pound;{:0,.2f}'.format(value)

    @staticmethod
    def nowrapfilter(text):
        return text.replace(' ', '&nbsp;')

    @staticmethod
    def get_output_filename(extension_type):
        if config.args.output_filename:
            filepath, file_extension = os.path.splitext(config.args.output_filename)
            if file_extension != extension_type:
                filepath = filepath + '.' + extension_type
        else:
            filepath = ReportPdf.DEFAULT_FILENAME + '.' + extension_type

        if not os.path.exists(filepath):
            return filepath

        filepath, file_extension = os.path.splitext(filepath)
        i = 2
        new_fname = '{}-{}{}'.format(filepath, i, file_extension)
        while os.path.exists(new_fname):
            i += 1
            new_fname = '{}-{}{}'.format(filepath, i, file_extension)

        return new_fname

class ReportLog(object):
    MAX_SYMBOL_LEN = 8
    MAX_NAME_LEN = 32
    ASSET_WIDTH = MAX_SYMBOL_LEN + MAX_NAME_LEN + 3

    def __init__(self, audit, tax_report, price_report, holdings_report):
        self.audit_report = audit
        self.tax_report = tax_report
        self.price_report = price_report
        self.holdings_report = holdings_report

        if config.args.taxyear:
            if self.audit and not config.args.summary:
                self.audit()
            log.info("==TAX YEAR %s/%s==", config.args.taxyear - 1, config.args.taxyear)
            self.capital_gains(config.args.taxyear)
            if not config.args.summary:
                self.income(config.args.taxyear)
                self.price_data(config.args.taxyear)
        else:
            if self.audit and not config.args.summary:
                self.audit()
            for tax_year in sorted(tax_report):
                log.info("==TAX YEAR %s/%s==", tax_year - 1, tax_year)
                self.capital_gains(tax_year)
                if not config.args.summary:
                    self.income(tax_year)

            if not config.args.summary:
                log.info("==APPENDIX==")
                for tax_year in sorted(tax_report):
                    self.price_data(tax_year)
                self.holdings()

    def audit(self):
        log.info("==FINAL AUDIT BALANCES==")
        for wallet in sorted(self.audit_report.wallets):
            for asset in sorted(self.audit_report.wallets[wallet]):
                log.info("%s:%s=%s",
                         wallet,
                         asset,
                         self.format_quantity(self.audit_report.wallets[wallet][asset]))

    def capital_gains(self, tax_year):
        cgains = self.tax_report[tax_year]['CapitalGains']

        log.info("--CAPITAL GAINS--")
        header = "%s %s %s %s %s %s %s %s" % ("Asset".ljust(self.MAX_SYMBOL_LEN),
                                              "Date".ljust(10),
                                              "Disposal Type".ljust(28),
                                              "Quantity".rjust(25),
                                              "Cost".rjust(13),
                                              "Fees".rjust(13),
                                              "Proceeds".rjust(13),
                                              "Gain".rjust(13))
        log.info(header)

        for asset in sorted(cgains.assets):
            for te in cgains.assets[asset]:
                log.info("%s %s %s %s %s %s %s %s",
                         te.asset.ljust(self.MAX_SYMBOL_LEN),
                         self.format_date(te.date),
                         te.format_disposal().ljust(28),
                         self.format_quantity(te.quantity).rjust(25),
                         self.format_value(te.cost).rjust(13),
                         self.format_value(te.fees).rjust(13),
                         self.format_value(te.proceeds).rjust(13),
                         self.format_value(te.gain).rjust(13))

        log.info("%s", '-' * len(header))
        log.info("%s %s %s %s %s %s %s %s",
                 "Total".ljust(self.MAX_SYMBOL_LEN),
                 ' ' * 10,
                 ' ' * 28,
                 ' ' * 25,
                 self.format_value(cgains.totals['cost']).rjust(13),
                 self.format_value(cgains.totals['fees']).rjust(13),
                 self.format_value(cgains.totals['proceeds']).rjust(13),
                 self.format_value(cgains.totals['gain']).rjust(13))

        log.info("-TAX SUMMARY-")
        log.info("Number of disposals=%s", cgains.summary['disposals'])
        log.info("Disposal proceeds=%s",
                 self.format_value(cgains.totals['proceeds']))

        if cgains.estimate['proceeds_warning']:
            log.warning("Assets sold are more than 4 times the annual allowance (%s), "
                        "this needs to be reported to HMRC",
                        self.format_value(cgains.estimate['allowance'] * 4))

        log.info("Allowable costs=%s",
                 self.format_value(cgains.totals['cost'] + cgains.totals['fees']))
        log.info("Gains in the year, before losses=%s",
                 self.format_value(cgains.summary['total_gain']))
        log.info("Losses in the year=%s",
                 self.format_value(abs(cgains.summary['total_loss'])))

        if not config.args.summary:
            log.info("-TAX ESTIMATE-")
            if cgains.totals['gain'] > 0:
                log.info("Taxable Gain=%s (%s of the tax-fee allowance %s used)",
                         self.format_value(cgains.estimate['taxable_gain']),
                         self.format_value(cgains.estimate['allowance_used']),
                         self.format_value(cgains.estimate['allowance']))
            else:
                log.info("Taxable Gain=%s",
                         self.format_value(cgains.estimate['taxable_gain']))
            log.info("Capital Gains Tax (Basic rate)=%s",
                     self.format_value(cgains.estimate['cgt_basic']))
            log.info("Capital Gains Tax (Higher rate)=%s",
                     self.format_value(cgains.estimate['cgt_higher']))

    def income(self, tax_year):
        income = self.tax_report[tax_year]['Income']

        log.info("--INCOME--")
        header = "%s %s %s %s %s %s" % ("Asset".ljust(self.MAX_SYMBOL_LEN),
                                        "Date".ljust(10),
                                        "Income Type".ljust(28),
                                        "Quantity".rjust(25),
                                        "Amount".rjust(13),
                                        "Fees".rjust(13))
        log.info(header)

        for asset in sorted(income.assets):
            for te in income.assets[asset]:
                log.info("%s %s %s %s %s %s",
                         te.asset.ljust(self.MAX_SYMBOL_LEN),
                         self.format_date(te.date),
                         te.type.ljust(28),
                         self.format_quantity(te.quantity).rjust(25),
                         self.format_value(te.amount).rjust(13),
                         self.format_value(te.fees).rjust(13))

        log.info("%s %s %s %s %s",
                 "Income Type".ljust(self.MAX_SYMBOL_LEN + 11),
                 ' ' * 28,
                 ' ' * 25,
                 "Amount".rjust(13),
                 "Fees".rjust(13))

        for i_type in sorted(income.type_totals):
            log.info("%s %s %s %s %s",
                     i_type.ljust(self.MAX_SYMBOL_LEN + 11),
                     ' ' * 28,
                     ' ' * 25,
                     self.format_value(income.type_totals[i_type]['amount']).rjust(13),
                     self.format_value(income.type_totals[i_type]['fees']).rjust(13))

        log.info("%s", '-' * len(header))
        log.info("%s %s %s %s %s",
                 "Total".ljust(self.MAX_SYMBOL_LEN + 11),
                 ' ' * 28,
                 ' ' * 25,
                 self.format_value(income.totals['amount']).rjust(13),
                 self.format_value(income.totals['fees']).rjust(13))

    def price_data(self, tax_year):
        log.info("--PRICE DATA %s/%s--", tax_year - 1, tax_year)
        log.info("%s %s %s %s %s",
                 "Asset".ljust(self.ASSET_WIDTH),
                 "Data Source".ljust(16),
                 "Date".ljust(10),
                 "Price (GBP)".rjust(13),
                 "Price (BTC)".rjust(25))

        if tax_year not in self.price_report:
            return

        for asset in sorted(self.price_report[tax_year]):
            for date in sorted(self.price_report[tax_year][asset]):
                price_data = self.price_report[tax_year][asset][date]
                if price_data['price_ccy'] is not None:
                    log.info("1 %s %s %s %s %s",
                             self.format_asset(asset, price_data['name']).ljust(self.ASSET_WIDTH),
                             price_data['data_source'].ljust(16),
                             self.format_date(date),
                             self.format_value(price_data['price_ccy']).rjust(13),
                             self.format_quantity(price_data['price_btc']).rjust(25))
                else:
                    log.info("1 %s %s %s %s",
                             self.format_asset(asset, price_data['name']).ljust(self.ASSET_WIDTH),
                             ' ' * 16,
                             self.format_date(date),
                             self.format_value(price_data['price_ccy']).rjust(13))

    def holdings(self):
        log.info("==CURRENT HOLDINGS==")
        header = "%s %s %s %s %s" % ("Asset".ljust(self.ASSET_WIDTH),
                                     "Quantity".rjust(25),
                                     "Cost + Fees".rjust(16),
                                     "Value".rjust(16),
                                     "Gain".rjust(16))
        log.info(header)

        for h in sorted(self.holdings_report['holdings']):
            holding = self.holdings_report['holdings'][h]
            if holding['value'] is not None:
                log.info("%s %s %s %s %s",
                         self.format_asset(holding['asset'],
                                           holding['name']).ljust(self.ASSET_WIDTH),
                         self.format_quantity(holding['quantity']).rjust(25),
                         self.format_value(holding['cost']).rjust(16),
                         self.format_value(holding['value']).rjust(16),
                         self.format_value(holding['gain']).rjust(16))
            else:
                log.info("%s %s %s %s",
                         self.format_asset(holding['asset'],
                                           holding['name']).ljust(self.ASSET_WIDTH),
                         self.format_quantity(holding['quantity']).rjust(25),
                         self.format_value(holding['cost']).rjust(16),
                         self.format_value(holding['value']).rjust(16))

        log.info("%s", '-' * len(header))
        log.info("%s %s %s %s %s",
                 "Total".ljust(self.ASSET_WIDTH),
                 ' ' * 25,
                 self.format_value(self.holdings_report['totals']['cost']).rjust(16),
                 self.format_value(self.holdings_report['totals']['value']).rjust(16),
                 self.format_value(self.holdings_report['totals']['gain']).rjust(16))

    @staticmethod
    def format_date(date):
        if isinstance(date, datetime):
            return date.strftime('%d/%m/%Y')
        else:
            return dateutil.parser.parse(date).strftime('%d/%m/%Y')

    @staticmethod
    def format_quantity(quantity):
        if quantity is not None:
            return '{:0,f}'.format(quantity.normalize())
        return 'n/a'

    @staticmethod
    def format_value(value):
        if value is not None:
            return config.sym() + '{:0,.2f}'.format(value + 0)
        return 'NOT AVAILABLE'

    @staticmethod
    def format_asset(asset, name):
        if name is not None:
            return '{} ({})'.format(asset, name)
        return asset



# -*- coding: utf-8 -*-
# Cryptoasset accounting, auditing and UK tax calculations (Capital Gains/Income Tax)
# (c) Nano Nano Ltd 2019

import logging
import argparse
import io
import sys
import codecs

import xlrd

from .version import __version__
from .config import config
from .import_records import ImportRecords
from .transactions import TransactionHistory
from .audit import AuditRecords
from .price.valueasset import ValueAsset
from .tax import TaxCalculator
from .report import ReportLog, ReportPdf

if sys.version_info[0] >= 3:
    sys.stdout.reconfigure(encoding='utf-8')

logging.basicConfig(stream=sys.stdout,
                    level=logging.INFO,
                    format='[%(asctime)s.%(msecs)03d] %(levelname)s -- : %(message)s',
                    datefmt='%Y-%m-%dT%H:%M:%S')
log = logging.getLogger()

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("filename",
                        type=str,
                        nargs='?',
                        help="filename of transaction records, "
                             "or can read CSV data from standard input")
    parser.add_argument("-v",
                        "--version",
                        action='version',
                        version='%(prog)s {version}'.format(version=__version__))
    parser.add_argument("-d",
                        "--debug",
                        action='store_true',
                        help="enabled debug logging")
    parser.add_argument("-ty",
                        "--taxyear",
                        type=int,
                        help="tax year")
    parser.add_argument("-s",
                        "--skipaudit",
                        action='store_true',
                        help="skip auditing of transactions")
    parser.add_argument("--summary",
                        action='store_true',
                        help="only output the capital gains summary in the tax report")
    parser.add_argument("-o",
                        dest='output_filename',
                        type=str,
                        help="specify the output filename for the tax report")
    parser.add_argument("--nopdf",
                        action='store_true',
                        help="don't output pdf report, output report to terminal only")

    config.args = parser.parse_args()
    config.args.nocache = False

    if config.args.debug:
        log.setLevel(logging.DEBUG)
        logging.getLogger('urllib3').setLevel(logging.WARNING)
        logging.getLogger('xhtml2pdf').setLevel(logging.WARNING)
        config.output_config(parser.prog)

    import_records = ImportRecords()

    if config.args.filename:
        try:
            import_records.import_excel(config.args.filename)
        except xlrd.XLRDError:
            with io.open(config.args.filename, newline='', encoding='utf-8') as csv_file:
                import_records.import_csv(csv_file)
    else:
        if sys.version_info[0] < 3:
            import_records.import_csv(codecs.getreader('utf-8')(sys.stdin))
        else:
            import_records.import_csv(sys.stdin)

    if import_records.failures():
        parser.exit()

    transaction_records = import_records.get_records()

    if not config.args.skipaudit and not config.args.summary:
        audit = AuditRecords(transaction_records)
    else:
        audit = None

    value_asset = ValueAsset()
    transaction_history = TransactionHistory(transaction_records, value_asset)

    tax = TaxCalculator(transaction_history.transactions)
    tax.pool_same_day()
    tax.match(tax.DISPOSAL_SAME_DAY)
    tax.match(tax.DISPOSAL_BED_AND_BREAKFAST)

    if config.args.debug:
        tax.output_transactions()

    tax.process_unmatched()

    if not config.args.summary:
        tax.process_income()

    if config.args.taxyear:
        tax.calculate_capital_gains(config.args.taxyear)
        if not config.args.summary:
            tax.calculate_income(config.args.taxyear)
    else:
        # Calculate for all years
        for year in sorted(tax.tax_events):
            tax.calculate_capital_gains(year)
            if not config.args.summary:
                tax.calculate_income(year)

        if not config.args.summary:
            tax.calculate_holdings(value_asset)

    if config.args.nopdf:
        ReportLog(audit,
                  tax.tax_report,
                  value_asset.price_report,
                  tax.holdings_report)
    else:
        ReportPdf(parser.prog,
                  audit,
                  tax.tax_report,
                  value_asset.price_report,
                  tax.holdings_report)



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

import logging
import copy
from decimal import Decimal
from datetime import datetime, timedelta

from .config import config
from .transactions import Buy, Sell
from .holdings import Holdings

log = logging.getLogger()

PRECISION = Decimal('0.00')

def which_tax_year(timestamp):
    if timestamp > datetime(timestamp.year, 4, 5, tzinfo=config.TZ_LOCAL):
        tax_year = timestamp.year + 1
    else:
        tax_year = timestamp.year

    return tax_year

class TaxCalculator(object):
    DISPOSAL_SAME_DAY = 'Same Day'
    DISPOSAL_BED_AND_BREAKFAST = 'Bed & Breakfast'
    DISPOSAL_SECTION_104 = 'Section 104'

    INCOME_TYPES = (Buy.TYPE_MINING, Buy.TYPE_INCOME)

    def __init__(self, transactions):
        self.transactions = transactions
        self.buys_ordered = []
        self.sells_ordered = []
        self.other_transactions = []

        self.tax_events = {}
        self.holdings = {}

        self.tax_report = {}
        self.holdings_report = {}

    def pool_same_day(self):
        transactions = copy.deepcopy(self.transactions)
        buy_transactions = {}
        sell_transactions = {}

        log.info("==POOL SAME DAY TRANSACTIONS==")
        for t in transactions:
            if isinstance(t, Buy) and t.acquisition:
                if (t.asset, t.timestamp.date()) not in buy_transactions:
                    buy_transactions[(t.asset, t.timestamp.date())] = t
                else:
                    buy_transactions[(t.asset, t.timestamp.date())] += t
            elif isinstance(t, Sell) and t.disposal:
                if (t.asset, t.timestamp.date()) not in sell_transactions:
                    sell_transactions[(t.asset, t.timestamp.date())] = t
                else:
                    sell_transactions[(t.asset, t.timestamp.date())] += t
            else:
                self.other_transactions.append(t)

        self.buys_ordered = sorted(buy_transactions.values())
        self.sells_ordered = sorted(sell_transactions.values())

        if config.args.debug:
            for t in sorted(self.buys_ordered + self.sells_ordered + self.other_transactions):
                log.debug(t)
                if len(t.pooled) > 1:
                    for tp in t.pooled:
                        log.debug("  %s", tp)

        log.debug("Total Transactions(Pooled)=%s",
                  len(self.buys_ordered + self.sells_ordered + self.other_transactions))

    def match(self, rule):
        log.info("==MATCH %s TRANSACTIONS==", rule.upper())
        sell_index = buy_index = 0

        if not self.buys_ordered:
            return

        while sell_index < len(self.sells_ordered):
            s = self.sells_ordered[sell_index]
            b = self.buys_ordered[buy_index]

            if (not s.matched and not b.matched and s.asset == b.asset and
                    self._rule_match(s.timestamp, b.timestamp, rule)):
                if config.args.debug:
                    log.debug(s)
                    if rule == self.DISPOSAL_BED_AND_BREAKFAST:
                        log.debug("%s (%s, %s days)",
                                  b, rule, (b.timestamp.date() - s.timestamp.date()).days)
                    else:
                        log.debug("%s (%s)", b, rule)

                if b.quantity > s.quantity:
                    b_remainder = b.split_buy(s.quantity)
                    self.buys_ordered.insert(buy_index + 1, b_remainder)
                elif s.quantity > b.quantity:
                    s_remainder = s.split_sell(b.quantity)
                    self.sells_ordered.insert(sell_index + 1, s_remainder)

                s.matched = b.matched = True
                tax_event = TaxEventCapitalGains(rule, b, s, b.cost,
                                            (b.fee_value or Decimal(0)) +
                                            (s.fee_value or Decimal(0)))
                self.tax_events[self._which_tax_year(tax_event.date)].append(tax_event)

                # Find next sell
                sell_index += 1
                buy_index = 0
            else:
                buy_index += 1
                if buy_index >= len(self.buys_ordered):
                    sell_index += 1
                    buy_index = 0

    def output_transactions(self):
        log.debug("==UPDATED TRANSACTIONS==")
        for t in sorted(self.buys_ordered +
                        self.sells_ordered +
                        self.other_transactions):
            log.debug(t)

        log.debug("Total Transactions=%s",
                  len(self.sells_ordered + self.buys_ordered + self.other_transactions))

    def _rule_match(self, s_timestamp, b_timestamp, rule):
        if rule == self.DISPOSAL_SAME_DAY:
            return b_timestamp.date() == s_timestamp.date()
        elif rule == self.DISPOSAL_BED_AND_BREAKFAST:
            return (s_timestamp.date() < b_timestamp.date() and
                    b_timestamp.date() <= s_timestamp.date() + timedelta(days=30))
        else:
            raise Exception

    def process_unmatched(self):
        log.info("==PROCESS UNMATCHED TRANSACTIONS==")
        unmatched_transactions = sorted([t for t in self.buys_ordered +
                                         self.sells_ordered +
                                         self.other_transactions if t.matched is False])

        for t in unmatched_transactions:
            if config.args.debug:
                if isinstance(t, Sell) and t.disposal:
                    log.debug("%s (Disposal)", t)
                else:
                    log.debug(t)

            if isinstance(t, Buy):
                self._add_tokens(t)
            elif isinstance(t, Sell):
                self._subtract_tokens(t)

    def _add_tokens(self, t):
        if t.asset not in self.holdings:
            self.holdings[t.asset] = Holdings(t.asset)

        if not t.acquisition:
            if config.transfers_include:
                # !IMPORTANT! - Make sure no disposal event occurs between a Withdrawal and a
                #   Deposit (of the same asset) otherwise the average cost basis would be incorrect
                cost = fees = Decimal(0)
            else:
                return
        else:
            cost = t.cost
            fees = t.fee_value or Decimal(0)

        self.holdings[t.asset].add_tokens(t.quantity, cost, fees)

    def _subtract_tokens(self, t):
        if t.asset not in self.holdings:
            self.holdings[t.asset] = Holdings(t.asset)

        if not t.disposal:
            if config.transfers_include:
                cost = fees = Decimal(0)
            else:
                return
        else:
            if self.holdings[t.asset].quantity:
                cost = self.holdings[t.asset].cost * (t.quantity /
                                                      self.holdings[t.asset].quantity)
                fees = self.holdings[t.asset].fees * (t.quantity /
                                                      self.holdings[t.asset].quantity)
            else:
                # Should never happen, only if incorrect transaction records
                cost = fees = Decimal(0)

        self.holdings[t.asset].subtract_tokens(t.quantity, cost, fees)

        if t.disposal:
            tax_event = TaxEventCapitalGains(self.DISPOSAL_SECTION_104,
                                        None, t, cost, fees + (t.fee_value or Decimal(0)))
            self.tax_events[self._which_tax_year(tax_event.date)].append(tax_event)

    def process_income(self):
        log.info("==PROCESS INCOME==")
        for t in self.transactions:
            if t.t_type in self.INCOME_TYPES:
                tax_event = TaxEventIncome(t)
                self.tax_events[self._which_tax_year(tax_event.date)].append(tax_event)

    def calculate_capital_gains(self, tax_year):
        log.info("==CALCULATE CAPITAL GAINS %s/%s==", tax_year - 1, tax_year)
        self.tax_report[tax_year] = {}
        self.tax_report[tax_year]['CapitalGains'] = CalculateCapitalGains(tax_year)

        if tax_year in self.tax_events:
            for te in sorted(self.tax_events[tax_year]):
                if isinstance(te, TaxEventCapitalGains):
                    self.tax_report[tax_year]['CapitalGains'].tax_summary(te)

        self.tax_report[tax_year]['CapitalGains'].tax_estimate(tax_year)

    def calculate_income(self, tax_year):
        log.info("==CALCULATE INCOME %s/%s==", tax_year - 1, tax_year)
        self.tax_report[tax_year]['Income'] = CalculateIncome()

        if tax_year in self.tax_events:
            for te in sorted(self.tax_events[tax_year]):
                if isinstance(te, TaxEventIncome):
                    self.tax_report[tax_year]['Income'].totalise(te)

        self.tax_report[tax_year]['Income'].totals_by_type()

    def calculate_holdings(self, value_asset):
        log.info("==CALCULATE HOLDINGS==")
        holdings = {}
        totals = {'cost': Decimal(0),
                  'value': Decimal(0),
                  'gain': Decimal(0)}

        for h in self.holdings:
            if self.holdings[h].quantity > 0 or config.show_empty_wallets:
                holdings[h] = {}
                holdings[h]['asset'] = self.holdings[h].asset
                holdings[h]['quantity'] = self.holdings[h].quantity
                holdings[h]['cost'] = (self.holdings[h].cost +
                                       self.holdings[h].fees).quantize(PRECISION)

                value, name, data_source = value_asset.get_current_value(self.holdings[h].asset,
                                                                         self.holdings[h].quantity)
                holdings[h]['value'] = value.quantize(PRECISION) if value is not None else None
                holdings[h]['name'] = name
                holdings[h]['data_source'] = data_source

                if holdings[h]['value'] is not None:
                    holdings[h]['gain'] = value - holdings[h]['cost']
                    totals['value'] += holdings[h]['value']
                    totals['gain'] += holdings[h]['gain']

                totals['cost'] += holdings[h]['cost']

        self.holdings_report['holdings'] = holdings
        self.holdings_report['totals'] = totals

    def _which_tax_year(self, timestamp):
        tax_year = which_tax_year(timestamp)
        if tax_year not in self.tax_events:
            self.tax_events[tax_year] = []

        return tax_year

class TaxEvent(object):
    def __init__(self, date, asset):
        self.date = date
        self.asset = asset

    def __eq__(self, other):
        return self.date == other.date

    def __ne__(self, other):
        return not self == other

    def __lt__(self, other):
        return self.date < other.date

class TaxEventCapitalGains(TaxEvent):
    def __init__(self, disposal_type, b, s, cost, fees):
        super(TaxEventCapitalGains, self).__init__(s.timestamp, s.asset)
        self.disposal_type = disposal_type
        self.quantity = s.quantity
        self.cost = cost.quantize(PRECISION)
        self.fees = fees.quantize(PRECISION)
        self.proceeds = s.proceeds.quantize(PRECISION)
        self.gain = self.proceeds - self.cost - self.fees
        self.acquisition_date = b.timestamp if b else None

        log.debug(" Gain=%s%s (proceeds=%s%s - cost=%s%s - fees=%s%s)",
                  config.sym(), '{:0,.2f}'.format(self.gain),
                  config.sym(), '{:0,.2f}'.format(self.proceeds),
                  config.sym(), '{:0,.2f}'.format(self.cost),
                  config.sym(), '{:0,.2f}'.format(self.fees))

    def format_disposal(self):
        if self.disposal_type == TaxCalculator.DISPOSAL_BED_AND_BREAKFAST:
            return self.disposal_type + " (" + self.acquisition_date.strftime('%d/%m/%Y') + ")"

        return self.disposal_type

class TaxEventIncome(TaxEvent):
    def __init__(self, b):
        super(TaxEventIncome, self).__init__(b.timestamp, b.asset)
        self.type = b.t_type
        self.quantity = b.quantity
        self.amount = b.cost.quantize(PRECISION)
        if b.fee_value:
            self.fees = b.fee_value.quantize(PRECISION)
        else:
            self.fees = Decimal(0)

class CalculateCapitalGains(object):
    CG_DATA_INDIVIDUALS = {2010: {'allowance': 10100, 'basic_rate': 18, 'higher_rate': 18},
                           2011: {'allowance': 10100, 'basic_rate': 18, 'higher_rate': 28},
                           2012: {'allowance': 10600, 'basic_rate': 18, 'higher_rate': 28},
                           2013: {'allowance': 10600, 'basic_rate': 18, 'higher_rate': 28},
                           2014: {'allowance': 10900, 'basic_rate': 18, 'higher_rate': 28},
                           2015: {'allowance': 11000, 'basic_rate': 18, 'higher_rate': 28},
                           2016: {'allowance': 11100, 'basic_rate': 18, 'higher_rate': 28},
                           2017: {'allowance': 11100, 'basic_rate': 10, 'higher_rate': 20},
                           2018: {'allowance': 11300, 'basic_rate': 10, 'higher_rate': 20},
                           2019: {'allowance': 11700, 'basic_rate': 10, 'higher_rate': 20},
                           2020: {'allowance': 12000, 'basic_rate': 10, 'higher_rate': 20}}

    def __init__(self, tax_year):
        self.totals = {'cost': Decimal(0),
                       'fees': Decimal(0),
                       'proceeds': Decimal(0),
                       'gain': Decimal(0)}
        self.summary = {'disposals': Decimal(0),
                        'total_gain': Decimal(0),
                        'total_loss': Decimal(0)}
        self.estimate = {'allowance': Decimal(self.CG_DATA_INDIVIDUALS[tax_year]['allowance']),
                         'allowance_used': Decimal(0),
                         'taxable_gain': Decimal(0),
                         'cgt_basic': Decimal(0),
                         'cgt_higher': Decimal(0),
                         'proceeds_warning': False}
        self.assets = {}

    def tax_summary(self, te):
        self.summary['disposals'] += 1
        self.totals['cost'] += te.cost
        self.totals['fees'] += te.fees
        self.totals['proceeds'] += te.proceeds
        self.totals['gain'] += te.gain
        if te.gain >= 0:
            self.summary['total_gain'] += te.gain
        else:
            self.summary['total_loss'] += te.gain

        if te.asset not in self.assets:
            self.assets[te.asset] = []

        self.assets[te.asset].append(te)

    def tax_estimate(self, tax_year):
        if self.totals['gain'] > self.estimate['allowance']:
            self.estimate['allowance_used'] = self.estimate['allowance']
            self.estimate['taxable_gain'] = self.totals['gain'] - self.estimate['allowance']
            self.estimate['cgt_basic'] = self.estimate['taxable_gain'] * \
                                         self.CG_DATA_INDIVIDUALS[tax_year]['basic_rate'] / 100
            self.estimate['cgt_higher'] = self.estimate['taxable_gain'] * \
                                          self.CG_DATA_INDIVIDUALS[tax_year]['higher_rate'] /100
        elif self.totals['gain'] > 0:
            self.estimate['allowance_used'] = self.totals['gain']

        if self.totals['proceeds'] >= self.estimate['allowance'] * 4:
            self.estimate['proceeds_warning'] = True

class CalculateIncome(object):
    def __init__(self):
        self.totals = {'amount': Decimal(0),
                       'fees': Decimal(0)}
        self.assets = {}
        self.types = {}
        self.type_totals = {}

    def totalise(self, te):
        self.totals['amount'] += te.amount
        self.totals['fees'] += te.fees

        if te.asset not in self.assets:
            self.assets[te.asset] = []

        self.assets[te.asset].append(te)

        if te.type not in self.types:
            self.types[te.type] = []

        self.types[te.type].append(te)

    def totals_by_type(self):
        for income_type in self.types:
            for te in self.types[income_type]:
                if income_type not in self.type_totals:
                    self.type_totals[income_type] = {}
                    self.type_totals[income_type]['amount'] = te.amount
                    self.type_totals[income_type]['fees'] = te.fees
                else:
                    self.type_totals[income_type]['amount'] += te.amount
                    self.type_totals[income_type]['fees'] += te.fees



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

import logging

from decimal import Decimal

from .config import config

log = logging.getLogger()

class Holdings(object):
    def __init__(self, asset):
        self.asset = asset
        self.quantity = Decimal(0)
        self.cost = Decimal(0)
        self.fees = Decimal(0)

    def add_tokens(self, quantity, cost, fees):
        self.quantity += quantity
        self.cost += cost
        self.fees += fees

        if config.args.debug:
            log.debug("%s=%s (+%s) cost=%s%s %s (+%s%s %s) fees=%s%s %s (+%s%s %s)",
                      self.asset,
                      '{:0,f}'.format(self.quantity.normalize()),
                      '{:0,f}'.format(quantity.normalize()),
                      config.sym(), '{:0,.2f}'.format(self.cost), config.CCY,
                      config.sym(), '{:0,.2f}'.format(cost), config.CCY,
                      config.sym(), '{:0,.2f}'.format(self.fees), config.CCY,
                      config.sym(), '{:0,.2f}'.format(fees), config.CCY)

    def subtract_tokens(self, quantity, cost, fees):
        self.quantity -= quantity
        self.cost -= cost
        self.fees -= fees

        if config.args.debug:
            log.debug("%s=%s (-%s) cost=%s%s %s (-%s%s %s) fees=%s%s %s (-%s%s %s)",
                      self.asset,
                      '{:0,f}'.format(self.quantity.normalize()),
                      '{:0,f}'.format(quantity.normalize()),
                      config.sym(), '{:0,.2f}'.format(self.cost), config.CCY,
                      config.sym(), '{:0,.2f}'.format(cost), config.CCY,
                      config.sym(), '{:0,.2f}'.format(self.fees), config.CCY,
                      config.sym(), '{:0,.2f}'.format(fees), config.CCY)



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

import logging
import sys
import platform
import re

import xlsxwriter
from xlsxwriter.utility import xl_rowcol_to_cell

from ..version import __version__
from ..config import config
from .out_record import TransactionOutRecord
from .output_csv import OutputBase

log = logging.getLogger()

if platform.system() == "Darwin":
    # Default size for MacOS
    FONT_SIZE = 12
else:
    FONT_SIZE = 11

class OutputExcel(OutputBase):
    FILE_EXTENSION = 'xlsx'
    DATE_FORMAT = 'yyyy-mm-dd hh:mm:ss'
    FONT_COLOR_IN_DATA = '#808080'
    TITLE = 'BittyTax Records'
    PROJECT_URL = 'https://github.com/BittyTax/BittyTax'

    def __init__(self, progname, data_files):
        super(OutputExcel, self).__init__(data_files)
        self.filename = self.get_output_filename(self.FILE_EXTENSION)
        self.workbook = xlsxwriter.Workbook(self.filename)
        self.workbook.set_size(1800, 1200)
        self.workbook.set_properties({'title': self.TITLE,
                                      'author': '{} {}'.format(progname, __version__),
                                      'comments': self.PROJECT_URL})

        self.format_out_header = self.workbook.add_format({'font_size': FONT_SIZE,
                                                           'font_color': 'white',
                                                           'bold': True,
                                                           'bg_color': 'black'})
        self.format_in_header = self.workbook.add_format({'font_size': FONT_SIZE,
                                                          'font_color': 'white',
                                                          'bold': True,
                                                          'bg_color': self.FONT_COLOR_IN_DATA})
        self.format_out_data = self.workbook.add_format({'font_size': FONT_SIZE,
                                                         'font_color': 'black'})
        self.format_in_data = self.workbook.add_format({'font_size': FONT_SIZE,
                                                        'font_color': self.FONT_COLOR_IN_DATA})
        self.format_in_data_err = self.workbook.add_format({'font_size': FONT_SIZE,
                                                            'font_color': self.FONT_COLOR_IN_DATA,
                                                            'diag_type': 3,
                                                            'diag_border': 7,
                                                            'diag_color': 'red'})
        self.format_num_float = self.workbook.add_format({'font_size': FONT_SIZE,
                                                          'font_color': 'black',
                                                          'num_format': '#,##0.' + '#' * 30})
        self.format_num_int = self.workbook.add_format({'num_format': '#,##0'})
        self.format_num_string = self.workbook.add_format({'font_size': FONT_SIZE,
                                                           'font_color': 'black',
                                                           'align': 'right'})
        self.format_currency = self.workbook.add_format({'font_size': FONT_SIZE,
                                                         'font_color': 'black',
                                                         'num_format': config.sym() + '#,##0.00'})
        self.format_timestamp = self.workbook.add_format({'font_size': FONT_SIZE,
                                                          'font_color': 'black',
                                                          'num_format': self.DATE_FORMAT})

    def write_excel(self):
        data_files = sorted(self.data_files, key=lambda df: df.parser.worksheet_name, reverse=False)
        for data_file in data_files:
            worksheet = Worksheet(self, data_file)

            data_rows = sorted(data_file.data_rows, key=lambda dr: dr.timestamp, reverse=False)
            for i, data_row in enumerate(data_rows):
                worksheet.add_row(data_row, i + 1)

            worksheet.autofit()

        self.workbook.close()
        log.info("Output EXCEL file created: %s", self.filename)

class Worksheet(object):
    BUY_LIST = (TransactionOutRecord.TYPE_DEPOSIT,
                TransactionOutRecord.TYPE_MINING,
                TransactionOutRecord.TYPE_INCOME,
                TransactionOutRecord.TYPE_GIFT_RECEIVED)

    SELL_LIST = (TransactionOutRecord.TYPE_WITHDRAWAL,
                 TransactionOutRecord.TYPE_SPEND,
                 TransactionOutRecord.TYPE_GIFT_SENT,
                 TransactionOutRecord.TYPE_CHARITY_SENT)
    SHEETNAME_MAX_LEN = 31

    names = {}

    def __init__(self, output, data_file):
        self.output = output
        self.worksheet = output.workbook.add_worksheet(self._sheet_name( \
                                                       data_file.parser.worksheet_name))
        self.col_width = {}

        #Add headings row
        self.worksheet.freeze_panes(1, len(self.output.BITTYTAX_OUT_HEADER))

        for col_num, col_name in enumerate(self.output.BITTYTAX_OUT_HEADER +
                                           data_file.parser.in_header):
            if col_num < len(self.output.BITTYTAX_OUT_HEADER):
                self.worksheet.write(0, col_num, col_name, self.output.format_out_header)
            else:
                self.worksheet.write(0, col_num, col_name, self.output.format_in_header)

            self._autofit_calc(col_num, len(col_name))

    def _sheet_name(self, parser_name):
        # Remove special characters
        name = re.sub(r'[/\\\?\*\[\]:]', '', parser_name)
        name = name[:self.SHEETNAME_MAX_LEN] if len(name) > self.SHEETNAME_MAX_LEN else name

        if name.lower() not in self.names:
            self.names[name.lower()] = 1
            sheet_name = name
        else:
            self.names[name.lower()] += 1
            sheet_name = '{}({})'.format(name, self.names[name.lower()])
            if len(sheet_name) > self.SHEETNAME_MAX_LEN:
                sheet_name = '{}({})'.format(name[:len(name) -
                                                  (len(sheet_name) - self.SHEETNAME_MAX_LEN)],
                                             self.names[name.lower()])

        return sheet_name

    def add_row(self, data_row, row_num):
        self.worksheet.set_row(row_num, None, self.output.format_out_data)

        # Add transaction record
        if data_row.t_record:
            self._xl_type(data_row.t_record.t_type, row_num, 0)
            self._xl_quantity(data_row.t_record.buy_quantity, row_num, 1)
            self._xl_asset(data_row.t_record.buy_asset, row_num, 2)
            self._xl_value(data_row.t_record.buy_value, row_num, 3)
            self._xl_quantity(data_row.t_record.sell_quantity, row_num, 4)
            self._xl_asset(data_row.t_record.sell_asset, row_num, 5)
            self._xl_value(data_row.t_record.sell_value, row_num, 6)
            self._xl_quantity(data_row.t_record.fee_quantity, row_num, 7)
            self._xl_asset(data_row.t_record.fee_asset, row_num, 8)
            self._xl_value(data_row.t_record.fee_value, row_num, 9)
            self._xl_wallet(data_row.t_record.wallet, row_num, 10)
            self._xl_timestamp(data_row.t_record.timestamp, row_num, 11)

        if sys.version_info[0] < 3:
            in_row = [r.decode('utf8') for r in data_row.in_row]
        else:
            in_row = data_row.in_row

        # Add original data
        for col_num, col_data in enumerate(in_row):
            if data_row.failure and data_row.failure.col_num == col_num:
                self.worksheet.write(row_num, 12 + col_num, col_data,
                                     self.output.format_in_data_err)
            else:
                self.worksheet.write(row_num, 12 + col_num, col_data,
                                     self.output.format_in_data)

            self._autofit_calc(12 + col_num, len(col_data))

    def _xl_type(self, t_type, row_num, col_num):
        if t_type in self.BUY_LIST:
            self.worksheet.data_validation(row_num, col_num, row_num, col_num,
                                           {'validate': 'list',
                                            'source': list(self.BUY_LIST)})
        elif t_type in self.SELL_LIST:
            self.worksheet.data_validation(row_num, col_num, row_num, col_num,
                                           {'validate': 'list',
                                            'source': list(self.SELL_LIST)})
        else:
            self.worksheet.data_validation(row_num, col_num, row_num, col_num,
                                           {'validate': 'list',
                                            'source': [TransactionOutRecord.TYPE_TRADE]})

        self.worksheet.write_string(row_num, col_num, t_type)
        self._autofit_calc(col_num, len(t_type))

    def _xl_quantity(self, quantity, row_num, col_num):
        if quantity is not None:
            if len(quantity.normalize().as_tuple().digits) > OutputBase.EXCEL_PRECISION:
                self.worksheet.write_string(row_num, col_num,
                                            '{0:f}'.format(quantity.normalize()),
                                            self.output.format_num_string)
            else:
                self.worksheet.write_number(row_num, col_num,
                                            quantity.normalize(), self.output.format_num_float)
                cell = xl_rowcol_to_cell(row_num, col_num)
                self.worksheet.conditional_format(row_num, col_num, row_num, col_num,
                                                  {'type': 'formula',
                                                   'criteria': '=INT(' + cell + ')=' + cell,
                                                   'format':  self.output.format_num_int})
            self._autofit_calc(col_num, len('{:0,f}'.format(quantity.normalize())))

    def _xl_asset(self, asset, row_num, col_num):
        self.worksheet.write_string(row_num, col_num, asset)
        self._autofit_calc(col_num, len(asset))

    def _xl_value(self, value, row_num, col_num):
        if value is not None:
            self.worksheet.write_number(row_num, col_num, value.normalize(),
                                        self.output.format_currency)
            self._autofit_calc(col_num, len('£{:0,.2f}'.format(value)))
        else:
            self.worksheet.write_blank(row_num, col_num, None, self.output.format_currency)

    def _xl_wallet(self, wallet, row_num, col_num):
        self.worksheet.write_string(row_num, col_num, wallet)
        self._autofit_calc(col_num, len(wallet))

    def _xl_timestamp(self, timestamp, row_num, col_num):
        utc_timestamp = timestamp.astimezone(config.TZ_UTC)
        utc_timestamp = timestamp.replace(tzinfo=None)

        self.worksheet.write_datetime(row_num, col_num, utc_timestamp, self.output.format_timestamp)
        self._autofit_calc(col_num, len(self.output.DATE_FORMAT))

    def _autofit_calc(self, col_num, width):
        if col_num in self.col_width:
            if width > self.col_width[col_num]:
                self.col_width[col_num] = width
        else:
            self.col_width[col_num] = width

    def autofit(self):
        for col_num in self.col_width:
            self.worksheet.set_column(col_num, col_num, self.col_width[col_num])



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

import logging
import csv
import sys
import io

import xlrd

from ..config import config
from .dataparser import DataParser
from .datarow import DataRow
from .exceptions import DataFormatUnrecognised

log = logging.getLogger()

class DataFile(object):
    FORMAT_EXCEL = 'Excel'
    FORMAT_CSV = 'CSV'

    CSV_DELIMITERS = (',', ';')

    data_files = {}
    data_files_ordered = []

    def __init__(self, file_format, filename, parser, reader):
        self.parser = parser
        self.data_rows = [DataRow(line_num + 1, in_row) for line_num, in_row in enumerate(reader)]

        if parser.row_handler:
            for data_row in self.data_rows:
                data_row.parse(parser)
        else:
            # all rows handled together
            DataRow.parse_all(self.data_rows, parser)

        failures = [data_row for data_row in self.data_rows if data_row.failure is not None]
        if failures:
            log.warning("Parser failure for %s file: %s", file_format, filename)
            for data_row in failures:
                log.warning("Row[%s]: %s",
                            parser.in_header_row_num + data_row.line_num, data_row.failure)

    def __eq__(self, other):
        return (self.parser.row_handler, self.parser.all_handler) == \
               (other.parser.row_handler, other.parser.all_handler)

    def __hash__(self):
        return hash((self.parser.row_handler, self.parser.all_handler))

    def __iadd__(self, other):
        if len(other.parser.header) > len(self.parser.header):
            self.parser = other.parser

        if config.args.duplicates:
            self.data_rows += [data_row
                               for data_row in other.data_rows if data_row not in self.data_rows]
        else:
            self.data_rows += other.data_rows

        return self

    @classmethod
    def read_excel(cls, filename):
        workbook = xlrd.open_workbook(filename)
        log.debug("EXCEL")

        sheet = workbook.sheet_by_index(0)
        reader = cls.get_cell_values(sheet.get_rows(), workbook)
        parser = cls.get_parser(reader)

        if parser is not None:
            log.info("File: %s matched as \"%s\"", filename, parser.worksheet_name)
            data_file = DataFile(cls.FORMAT_EXCEL, filename, parser, reader)
            cls.consolidate_datafiles(data_file)
        else:
            raise DataFormatUnrecognised

        workbook.release_resources()
        del workbook

    @staticmethod
    def get_cell_values(rows, workbook):
        for row in rows:
            yield [DataFile.convert_cell(cell, workbook) for cell in row]

    @staticmethod
    def convert_cell(cell, workbook):
        if cell.ctype == xlrd.XL_CELL_DATE:
            value = xlrd.xldate.xldate_as_datetime(cell.value, workbook.datemode). \
                         strftime('%Y-%m-%dT%H:%M:%S %Z')
        elif cell.ctype == xlrd.XL_CELL_NUMBER:
            # repr is required to ensure no precision is lost
            value = repr(cell.value)
        else:
            value = str(cell.value)

        return value

    @classmethod
    def read_csv(cls, filename):
        with io.open(filename, newline='', encoding='utf-8-sig') as csv_file:
            for delimiter in cls.CSV_DELIMITERS:
                log.debug("CSV, delimiter='%s'", delimiter)

                if sys.version_info[0] < 3:
                    # special handling required for utf-8 encoded csv files
                    reader = csv.reader(cls.utf_8_encoder(csv_file), delimiter=delimiter)
                else:
                    reader = csv.reader(csv_file, delimiter=delimiter)

                parser = cls.get_parser(reader)
                if parser is not None:
                    log.info("File: %s matched as \"%s\"", filename, parser.worksheet_name)
                    data_file = DataFile(cls.FORMAT_CSV, filename, parser, reader)
                    cls.consolidate_datafiles(data_file)
                    break
                else:
                    csv_file.seek(0)

        if parser is None:
            raise DataFormatUnrecognised

    @classmethod
    def consolidate_datafiles(cls, data_file):
        if data_file in cls.data_files:
            cls.data_files[data_file] += data_file
        else:
            cls.data_files[data_file] = data_file
            cls.data_files_ordered.append(data_file)

    @staticmethod
    def utf_8_encoder(unicode_csv_data):
        for line in unicode_csv_data:
            yield line.encode('utf-8')

    @staticmethod
    def get_parser(reader):
        parser = None
        # header might not be on first line
        for row in range(5):
            try:
                parser = DataParser.match_header(next(reader), row)
            except KeyError:
                continue
            except StopIteration:
                pass
            except UnicodeDecodeError:
                break
            else:
                break

        return parser



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

import logging
import copy
from decimal import Decimal

from .config import config
from .record import TransactionRecord

log = logging.getLogger()

class TransactionHistory(object):
    def __init__(self, transaction_records, value_asset):
        self.value_asset = value_asset
        self.transactions = []

        log.info("==SPLIT TRANSACTION RECORDS==")

        for tr in transaction_records:
            log.debug(tr)
            self.get_all_values(tr)

            # Attribute the fee value to the buy, the sell or both
            if tr.fee and tr.fee.disposal and tr.fee.proceeds is not None:
                if tr.buy and tr.sell:
                    if tr.buy.asset in config.fiat_list:
                        tr.sell.fee_value = tr.fee.proceeds
                        tr.sell.fee_fixed = tr.fee.proceeds_fixed
                    elif tr.sell.asset in config.fiat_list:
                        tr.buy.fee_value = tr.fee.proceeds
                        tr.buy.fee_fixed = tr.fee.proceeds_fixed
                    else:
                        # Split fee between both
                        tr.buy.fee_value = tr.fee.proceeds / 2
                        tr.buy.fee_fixed = tr.fee.proceeds_fixed
                        tr.sell.fee_value = tr.fee.proceeds - tr.buy.fee_value
                        tr.sell.fee_fixed = tr.fee.proceeds_fixed
                elif tr.buy:
                    tr.buy.fee_value = tr.fee.proceeds
                    tr.buy.fee_fixed = tr.fee.proceeds_fixed
                elif tr.sell:
                    tr.sell.fee_value = tr.fee.proceeds
                    tr.sell.fee_fixed = tr.fee.proceeds_fixed

            if tr.buy and tr.buy.asset not in config.fiat_list:
                tr.buy.set_tid()
                self.transactions.append(tr.buy)
                log.debug(tr.buy)
            if tr.sell and tr.sell.asset not in config.fiat_list:
                tr.sell.set_tid()
                self.transactions.append(tr.sell)
                log.debug(tr.sell)
            if tr.fee and tr.fee.asset not in config.fiat_list:
                tr.fee.set_tid()
                self.transactions.append(tr.fee)
                log.debug(tr.fee)

        log.info("Total transactions=%s", len(self.transactions))

    def get_all_values(self, tr):
        if tr.buy and tr.buy.acquisition and tr.buy.cost is None:
            if tr.sell:
                (tr.buy.cost,
                 tr.buy.cost_fixed) = self.which_asset_value(tr)
            else:
                (tr.buy.cost,
                 tr.buy.cost_fixed) = self.value_asset.get_value(tr.buy.asset,
                                                                 tr.buy.timestamp,
                                                                 tr.buy.quantity)

        if tr.sell and tr.sell.disposal and tr.sell.proceeds is None:
            if tr.buy:
                tr.sell.proceeds = tr.buy.cost
                tr.sell.proceeds_fixed = tr.buy.cost_fixed
            else:
                (tr.sell.proceeds,
                 tr.sell.proceeds_fixed) = self.value_asset.get_value(tr.sell.asset,
                                                                      tr.sell.timestamp,
                                                                      tr.sell.quantity)
        if tr.fee and tr.fee.disposal and tr.fee.proceeds is None:
            if tr.fee.asset not in config.fiat_list:
                if tr.buy and tr.buy.asset == tr.fee.asset:
                    price = tr.buy.cost / tr.buy.quantity if tr.buy.cost else Decimal(0)
                    tr.fee.proceeds = tr.fee.quantity * price
                    tr.fee.proceeds_fixed = tr.buy.cost_fixed
                elif tr.sell and tr.sell.asset == tr.fee.asset:
                    price = tr.sell.proceeds / tr.sell.quantity if tr.sell.proceeds else Decimal(0)
                    tr.fee.proceeds = tr.fee.quantity * price
                    tr.fee.proceeds_fixed = tr.sell.proceeds_fixed
                else:
                    # Must be a 3rd cryptoasset
                    (tr.fee.proceeds,
                     tr.fee.proceeds_fixed) = self.value_asset.get_value(tr.fee.asset,
                                                                         tr.fee.timestamp,
                                                                         tr.fee.quantity)
            else:
                # Fee paid in fiat
                (tr.fee.proceeds,
                 tr.fee.proceeds_fixed) = self.value_asset.get_value(tr.fee.asset,
                                                                     tr.fee.timestamp,
                                                                     tr.fee.quantity)
    def which_asset_value(self, tr):
        if config.trade_asset_type == config.TRADE_ASSET_TYPE_BUY:
            value, fixed = self.value_asset.get_value(tr.buy.asset,
                                                      tr.buy.timestamp,
                                                      tr.buy.quantity)
        elif config.trade_asset_type == config.TRADE_ASSET_TYPE_SELL:
            value, fixed = self.value_asset.get_value(tr.sell.asset,
                                                      tr.sell.timestamp,
                                                      tr.sell.quantity)
        else:
            pos_sell_asset = pos_buy_asset = len(config.asset_priority) + 1

            if tr.sell.asset in config.asset_priority:
                pos_sell_asset = config.asset_priority.index(tr.sell.asset)
            if tr.buy.asset in config.asset_priority:
                pos_buy_asset = config.asset_priority.index(tr.buy.asset)

            if pos_sell_asset <= pos_buy_asset:
                value, fixed = self.value_asset.get_value(tr.sell.asset,
                                                          tr.sell.timestamp,
                                                          tr.sell.quantity)
            else:
                value, fixed = self.value_asset.get_value(tr.buy.asset,
                                                          tr.buy.timestamp,
                                                          tr.buy.quantity)
        return value, fixed

class TransactionBase(object):
    def __init__(self, t_type, asset, quantity):
        self.tid = None
        self.t_record = None
        self.t_type = t_type
        self.asset = asset
        self.quantity = quantity
        self.fee_value = None
        self.fee_fixed = True
        self.wallet = None
        self.timestamp = None
        self.matched = False
        self.pooled = []

    def set_tid(self):
        self.tid = self.t_record.set_tid()

    def _format_tid(self):
        return str(self.tid[0]) + "." + str(self.tid[1])

    def _format_quantity(self):
        if self.quantity is None:
            return '-'
        return '{:0,f}'.format(self.quantity.normalize())

    def _format_match_status(self):
        return " (M)" if self.matched else ""

    def _format_pooled(self):
        if self.pooled:
            return " [" + str(len(self.pooled)) + "]"
        return ''

    def _format_fee(self):
        if self.fee_value is not None:
            if self.fee_fixed:
                return ' + Fee=' + config.sym() + '{:0,.2f} {}'.format(self.fee_value, config.CCY)
            else:
                return ' + Fee=~' + config.sym() + '{:0,.2f} {}'.format(self.fee_value, config.CCY)

        return ''

    def __eq__(self, other):
        return (self.asset, self.timestamp) == (other.asset, other.timestamp)

    def __ne__(self, other):
        return not self == other

    def __lt__(self, other):
        return (self.asset, self.timestamp) < (other.asset, other.timestamp)

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k == 't_record':
                # Keep reference to the transaction record
                setattr(result, k, v)
            else:
                setattr(result, k, copy.deepcopy(v, memo))
        return result

class Buy(TransactionBase):
    TYPE_DEPOSIT = TransactionRecord.TYPE_DEPOSIT
    TYPE_MINING = TransactionRecord.TYPE_MINING
    TYPE_INCOME = TransactionRecord.TYPE_INCOME
    TYPE_GIFT_RECEIVED = TransactionRecord.TYPE_GIFT_RECEIVED
    TYPE_TRADE = TransactionRecord.TYPE_TRADE

    ACQUISITION_TYPES = {TYPE_MINING, TYPE_INCOME, TYPE_GIFT_RECEIVED, TYPE_TRADE}

    def __init__(self, t_type, buy_quantity, buy_asset, buy_value):
        super(Buy, self).__init__(t_type, buy_asset, buy_quantity)
        self.cost = buy_value
        if self.cost is not None:
            self.cost_fixed = True
        else:
            self.cost_fixed = False

        self.acquisition = bool(self.t_type in self.ACQUISITION_TYPES)

    def __iadd__(self, other):
        if not self.pooled:
            self.pooled.append(copy.deepcopy(self))

        # Pool buys
        if self.asset != other.asset:
            raise Exception
        self.quantity += other.quantity
        self.cost += other.cost

        if self.fee_value is not None and other.fee_value is not None:
            self.fee_value += other.fee_value
        elif self.fee_value is None and other.fee_value is not None:
            self.fee_value = other.fee_value

        if other.timestamp < self.timestamp:
            # Keep timestamp of earliest transaction
            self.timestamp = other.timestamp

        if other.wallet != self.wallet:
            self.wallet = "<pooled>"

        if other.cost_fixed != self.cost_fixed:
            self.cost_fixed = False

        self.pooled.append(other)
        return self

    def split_buy(self, sell_quantity):
        remainder = copy.deepcopy(self)

        self.cost = self.cost * (sell_quantity / self.quantity)

        if self.fee_value:
            self.fee_value = self.fee_value * (sell_quantity / self.quantity)

        self.quantity = sell_quantity
        self.set_tid()

        remainder.cost = remainder.cost - self.cost

        if self.fee_value:
            remainder.fee_value = remainder.fee_value - self.fee_value

        remainder.quantity = remainder.quantity - sell_quantity
        remainder.set_tid()

        log.debug("split: %s", self)
        log.debug("split: %s", remainder)

        return remainder

    def _format_cost(self):
        if self.cost is not None:
            if self.cost_fixed:
                return ' (=' + config.sym() + '{:0,.2f} {})'.format(self.cost, config.CCY)
            else:
                return ' (~' + config.sym() + '{:0,.2f} {})'.format(self.cost, config.CCY)
        return ''

    def _format_acquisition(self):
        if not self.acquisition:
            return '*'
        return ''

    def __str__(self):
        return "T-" + \
               type(self).__name__ + \
               self._format_acquisition() + " " + \
               "[TID:" + self._format_tid() + "] " + \
               self.t_type + ": " + \
               self._format_quantity() + " " + \
               self.asset + \
               self._format_cost() + \
               self._format_fee() + " '" + \
               self.wallet + "' " + \
               self.timestamp.strftime('%Y-%m-%dT%H:%M:%S %Z') + \
               self._format_pooled() + \
               self._format_match_status()

class Sell(TransactionBase):
    TYPE_WITHDRAWAL = TransactionRecord.TYPE_WITHDRAWAL
    TYPE_SPEND = TransactionRecord.TYPE_SPEND
    TYPE_GIFT_SENT = TransactionRecord.TYPE_GIFT_SENT
    TYPE_CHARITY_SENT = TransactionRecord.TYPE_CHARITY_SENT
    TYPE_TRADE = TransactionRecord.TYPE_TRADE

    DISPOSAL_TYPES = {TYPE_SPEND, TYPE_GIFT_SENT, TYPE_CHARITY_SENT, TYPE_TRADE}

    def __init__(self, t_type, sell_quantity, sell_asset, sell_value):
        super(Sell, self).__init__(t_type, sell_asset, sell_quantity)
        self.proceeds = sell_value
        if self.proceeds is not None:
            self.proceeds_fixed = True
        else:
            self.proceeds_fixed = False

        self.disposal = bool(self.t_type in self.DISPOSAL_TYPES)

    def __iadd__(self, other):
        if not self.pooled:
            self.pooled.append(copy.deepcopy(self))

        # Pool sells
        if self.asset != other.asset:
            raise Exception
        self.quantity += other.quantity
        self.proceeds += other.proceeds

        if self.fee_value is not None and other.fee_value is not None:
            self.fee_value += other.fee_value
        elif self.fee_value is None and other.fee_value is not None:
            self.fee_value = other.fee_value

        if other.timestamp > self.timestamp:
            # Keep timestamp of latest transaction
            self.timestamp = other.timestamp

        if other.wallet != self.wallet:
            self.wallet = "<pooled>"

        if other.proceeds_fixed != self.proceeds_fixed:
            self.proceeds_fixed = False

        self.pooled.append(other)
        return self

    def split_sell(self, buy_quantity):
        remainder = copy.deepcopy(self)

        self.proceeds = self.proceeds * (buy_quantity / self.quantity)

        if self.fee_value:
            self.fee_value = self.fee_value * (buy_quantity / self.quantity)

        self.quantity = buy_quantity
        self.set_tid()

        remainder.proceeds = remainder.proceeds - self.proceeds

        if self.fee_value:
            remainder.fee_value = remainder.fee_value - self.fee_value

        remainder.quantity = remainder.quantity - buy_quantity
        remainder.set_tid()

        log.debug("split: %s", self)
        log.debug("split: %s", remainder)

        return remainder

    def _format_proceeds(self):
        if self.proceeds is not None:
            if self.proceeds_fixed:
                return ' (=' + config.sym() + '{:0,.2f} {})'.format(self.proceeds, config.CCY)
            else:
                return ' (~' + config.sym() + '{:0,.2f} {})'.format(self.proceeds, config.CCY)
        return ''

    def _format_disposal(self):
        if not self.disposal:
            return '*'
        return ''

    def __str__(self):
        return "T-" + \
               type(self).__name__ + \
               self._format_disposal() + " " + \
               "[TID:" + self._format_tid() + "] " + \
               self.t_type + ": " + \
               self._format_quantity() + " " + \
               self.asset + \
               self._format_proceeds() + \
               self._format_fee() + " '" + \
               self.wallet + "' " + \
               self.timestamp.strftime('%Y-%m-%dT%H:%M:%S %Z') + \
               self._format_pooled() + \
               self._format_match_status()



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

import logging
import argparse
import sys

import xlrd

from ..version import __version__
from ..config import config
from .dataparser import DataParser
from .datafile import DataFile
from .output_csv import OutputCsv
from .output_excel import OutputExcel
from .exceptions import UnknownCryptoassetError, UnknownUsernameError, DataFormatUnrecognised

if sys.version_info[0] >= 3:
    sys.stderr.reconfigure(encoding='utf-8')

logging.basicConfig(stream=sys.stderr,
                    level=logging.INFO,
                    format='[%(asctime)s.%(msecs)03d] %(levelname)s -- : %(message)s',
                    datefmt='%Y-%m-%dT%H:%M:%S')
log = logging.getLogger()

def main():
    parser = argparse.ArgumentParser(epilog="supported data file formats:\n" + \
                                     DataParser.format_parsers(),
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument("filename",
                        type=str,
                        nargs='+',
                        help="filename of data file")
    parser.add_argument("-v",
                        "--version",
                        action='version',
                        version='%(prog)s {version}'.format(version=__version__))
    parser.add_argument("-d",
                        "--debug",
                        action='store_true',
                        help="enabled debug logging")
    parser.add_argument("-uc",
                        "--unconfirmed",
                        action='store_true',
                        help="include unconfirmed transactions")
    parser.add_argument("-ca",
                        dest="cryptoasset",
                        type=str,
                        help="specify a cryptoasset symbol, if it cannot be identified "
                             "automatically")
    parser.add_argument("--duplicates",
                        action='store_true',
                        help="remove any duplicate input rows across data files")
    parser.add_argument("--format",
                        choices=[config.FORMAT_EXCEL, config.FORMAT_CSV, config.FORMAT_RECAP],
                        default=config.FORMAT_EXCEL,
                        type=str.upper,
                        help="specify the output format")
    parser.add_argument("-nh",
                        "--noheader",
                        action='store_true',
                        help="exclude header from CSV output")
    parser.add_argument("-a",
                        "--append",
                        action='store_true',
                        help="append original data as extra columns in the CSV output")
    parser.add_argument("-s",
                        "--sort",
                        action='store_true',
                        help="sort CSV output by timestamp")
    parser.add_argument("-o",
                        dest='output_filename',
                        type=str,
                        help="specify the output filename")

    config.args = parser.parse_args()

    if config.args.debug:
        log.setLevel(logging.DEBUG)
        config.output_config(parser.prog)

    for filename in config.args.filename:
        try:
            try:
                DataFile.read_excel(filename)
            except xlrd.XLRDError:
                DataFile.read_csv(filename)
        except UnknownCryptoassetError:
            parser.error("cryptoasset cannot be identified for data file: {}, "
                         "please specify using the [-ca CRYPTOASSET] option".format(filename))
        except UnknownUsernameError:
            parser.exit("{}: error: username cannot be identified in data file: {}, "
                        "please specify usernames in the {} file"
                        .format(parser.prog, filename, config.BITTYTAX_CONFIG))
        except FileNotFoundError:
            log.warning("File not found: %s", filename)
        except IsADirectoryError:
            log.warning("File is a directory: %s", filename)
        except DataFormatUnrecognised:
            log.warning("Data file format unrecognised: %s", filename)

    if DataFile.data_files:
        if config.args.format == config.FORMAT_EXCEL:
            output = OutputExcel(parser.prog, DataFile.data_files_ordered)
            output.write_excel()
        else:
            output = OutputCsv(DataFile.data_files_ordered)
            output.write_csv()



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

class DataParserError(Exception):
    def __init__(self, col_num, col_name, value=None):
        super(DataParserError, self).__init__()
        self.col_num = col_num
        self.col_name = col_name
        self.value = value

class UnexpectedTypeError(DataParserError):
    def __str__(self):
        return 'Unrecognised {}: \'{}\''.format(self.col_name, self.value)

class MissingValueError(DataParserError):
    def __str__(self):
        return 'Missing value for \'{}\''.format(self.col_name)

class MissingComponentError(DataParserError):
    def __str__(self):
        return 'Missing component data for {}: \'{}\''.format(self.col_name, self.value)

class UnexpectedTradingPairError(DataParserError):
    def __str__(self):
        return 'Unrecognised trading pair for {}: \'{}\''.format(self.col_name, self.value)

class UnknownCryptoassetError(Exception):
    def __str__(self):
        return 'Cryptoasset cannot be identified'

class UnknownUsernameError(Exception):
    def __str__(self):
        return 'Username cannot be identified'

class DataFormatUnrecognised(Exception):
    def __str__(self):
        return 'Data file format unrecognised'



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019
from .record import TransactionRecord

class TransactionParserError(Exception):
    def __init__(self, col_num, col_name, value=None):
        super(TransactionParserError, self).__init__()
        self.col_num = col_num
        self.col_name = col_name
        self.value = value

class UnexpectedTransactionTypeError(TransactionParserError):
    def __str__(self):
        return 'Invalid Transaction Type: \'{}\', use {{{}}}'.format(self.value, \
                ','.join(TransactionRecord.ALL_TYPES))

class TimestampParserError(TransactionParserError):
    def __str__(self):
        return 'Invalid Timestamp: \'{}\', use format YYYY-MM-DDTHH:MM:SS ZZZ'.format(self.value)

class DataValueError(TransactionParserError):
    def __str__(self):
        return 'Invalid data for {}: \'{}\''.format(self.col_name, self.value)

class UnexpectedDataError(TransactionParserError):
    def __str__(self):
        return 'Unexpected data in {}: \'{}\''.format(self.col_name, self.value)

class MissingDataError(TransactionParserError):
    def __str__(self):
        return 'Missing data for {}'.format(self.col_name)



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

import datetime
import logging

from .parsers import *
from ..config import config
from .exceptions import UnknownCryptoassetError, DataParserError

DEFAULT_TIMESTAMP = datetime.datetime(datetime.MINYEAR, 1, 1, tzinfo=config.TZ_UTC)

log = logging.getLogger()

class DataRow(object):
    def __init__(self, line_num, in_row):
        self.line_num = line_num
        self.in_row = in_row
        self.timestamp = DEFAULT_TIMESTAMP
        self.t_record = None
        self.parsed = False
        self.failure = None

    def parse(self, parser):
        try:
            parser.row_handler(self, parser)
        except UnknownCryptoassetError:
            raise
        except DataParserError as e:
            self.failure = e

    def __eq__(self, other):
        return self.in_row == other.in_row

    def __hash__(self):
        return hash(self.in_row)

    @staticmethod
    def parse_all(data_rows, parser):
        parser.all_handler(data_rows, parser)



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

from decimal import Decimal

from ..record import TransactionRecord

class TransactionOutRecord(object):
    TYPE_DEPOSIT = TransactionRecord.TYPE_DEPOSIT
    TYPE_MINING = TransactionRecord.TYPE_MINING
    TYPE_INCOME = TransactionRecord.TYPE_INCOME
    TYPE_GIFT_RECEIVED = TransactionRecord.TYPE_GIFT_RECEIVED
    TYPE_WITHDRAWAL = TransactionRecord.TYPE_WITHDRAWAL
    TYPE_SPEND = TransactionRecord.TYPE_SPEND
    TYPE_GIFT_SENT = TransactionRecord.TYPE_GIFT_SENT
    TYPE_CHARITY_SENT = TransactionRecord.TYPE_CHARITY_SENT
    TYPE_TRADE = TransactionRecord.TYPE_TRADE

    def __init__(self, t_type, timestamp,
                 buy_quantity=None, buy_asset="", buy_value=None,
                 sell_quantity=None, sell_asset="", sell_value=None,
                 fee_quantity=None, fee_asset="", fee_value=None,
                 wallet=""):

        self.t_type = t_type
        self.buy_quantity = Decimal(buy_quantity) if buy_quantity is not None else None
        self.buy_asset = buy_asset
        self.buy_value = Decimal(buy_value) if buy_value is not None else None
        self.sell_quantity = Decimal(sell_quantity) if sell_quantity is not None else None
        self.sell_asset = sell_asset
        self.sell_value = Decimal(sell_value) if sell_value is not None else None
        self.fee_quantity = Decimal(fee_quantity) if fee_quantity is not None else None
        self.fee_asset = fee_asset
        self.fee_value = Decimal(fee_value) if fee_value is not None else None
        self.wallet = wallet
        self.timestamp = timestamp

    @staticmethod
    def format_quantity(quantity):
        if quantity is None:
            return '-'

        return '{:0,f}'.format(quantity.normalize())



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

from .config import config

class TransactionRecord(object):
    TYPE_DEPOSIT = 'Deposit'
    TYPE_MINING = 'Mining'
    TYPE_INCOME = 'Income'
    TYPE_GIFT_RECEIVED = 'Gift-Received'
    TYPE_WITHDRAWAL = 'Withdrawal'
    TYPE_SPEND = 'Spend'
    TYPE_GIFT_SENT = 'Gift-Sent'
    TYPE_CHARITY_SENT = 'Charity-Sent'
    TYPE_TRADE = 'Trade'

    ALL_TYPES = (TYPE_DEPOSIT,
                 TYPE_MINING,
                 TYPE_INCOME,
                 TYPE_GIFT_RECEIVED,
                 TYPE_WITHDRAWAL,
                 TYPE_SPEND,
                 TYPE_GIFT_SENT,
                 TYPE_CHARITY_SENT,
                 TYPE_TRADE)
    cnt = 0

    def __init__(self, t_type, buy, sell, fee, wallet, timestamp):
        self.tid = None
        self.t_type = t_type
        self.buy = buy
        self.sell = sell
        self.fee = fee
        self.wallet = wallet
        self.timestamp = timestamp

        if self.buy:
            self.buy.t_record = self
            self.buy.timestamp = self.timestamp.astimezone(config.TZ_LOCAL)
            self.buy.wallet = self.wallet
        if self.sell:
            self.sell.t_record = self
            self.sell.timestamp = self.timestamp.astimezone(config.TZ_LOCAL)
            self.sell.wallet = self.wallet
        if self.fee:
            self.fee.t_record = self
            self.fee.timestamp = self.timestamp.astimezone(config.TZ_LOCAL)
            self.fee.wallet = self.wallet

    def set_tid(self):
        if self.tid is None:
            TransactionRecord.cnt += 1
            self.tid = [TransactionRecord.cnt, 0]
        else:
            self.tid[1] += 1

        return list(self.tid)

    def _format_fee(self):
        if self.fee:
            return ' + Fee=' + self.format_quantity(self.fee.quantity) + \
                   ' ' + self.fee.asset + self.format_value(self.fee.proceeds)
        return ''

    @staticmethod
    def format_quantity(quantity):
        if quantity is None:
            return '-'

        return '{:0,f}'.format(quantity.normalize())

    @staticmethod
    def format_value(value):
        if value is not None:
            return ' (' + config.sym() + '{:0,.2f} {})'.format(value, config.CCY)

        return ''

    def __eq__(self, other):
        return self.timestamp == other.timestamp

    def __ne__(self, other):
        return not self == other

    def __lt__(self, other):
        return self.timestamp < other.timestamp

    def __str__(self):
        if self.buy and self.sell:
            return "TR [TID:" + str(self.tid[0]) + "] " + \
                   self.t_type + ": " + \
                   self.format_quantity(self.buy.quantity) + " " + \
                   self.buy.asset + \
                   self.format_value(self.buy.cost) + " <- " + \
                   self.format_quantity(self.sell.quantity) + " " + \
                   self.sell.asset + \
                   self.format_value(self.sell.proceeds) + \
                   self._format_fee() + " '" + \
                   self.wallet + "' " + \
                   self.timestamp.strftime('%Y-%m-%dT%H:%M:%S %Z')
        elif self.buy:
            return "TR [TID:" + str(self.tid[0]) + "] " + \
                   self.t_type + ": " + \
                   self.format_quantity(self.buy.quantity) + " " + \
                   self.buy.asset + \
                   self.format_value(self.buy.cost) + \
                   self._format_fee() + " '" + \
                   self.wallet + "' " + \
                   self.timestamp.strftime('%Y-%m-%dT%H:%M:%S %Z')
        elif self.sell:
            return "TR [TID:" + str(self.tid[0]) + "] " + \
                   self.t_type + ": " + \
                   self.format_quantity(self.sell.quantity) + " " + \
                   self.sell.asset + \
                   self.format_value(self.sell.proceeds) + \
                   self._format_fee() + " '" + \
                   self.wallet + "' " + \
                   self.timestamp.strftime('%Y-%m-%dT%H:%M:%S %Z')

        return ''



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

import logging
import sys
import csv
from decimal import Decimal, InvalidOperation

import dateutil.parser
import xlrd

from .config import config
from .transactions import Buy, Sell
from .record import TransactionRecord
from .exceptions import TransactionParserError, UnexpectedTransactionTypeError, \
                        TimestampParserError, DataValueError, MissingDataError, \
                        UnexpectedDataError

log = logging.getLogger()

class ImportRecords(object):
    def __init__(self):
        self.t_rows = []

    def import_excel(self, filename):
        workbook = xlrd.open_workbook(filename)
        log.info("==IMPORT TRANSACTION RECORDS FROM EXCEL FILE: %s ==", filename)

        for worksheet in workbook.sheets():
            for row_num in range(1, worksheet.nrows):
                row = [self.convert_cell(worksheet.cell(row_num, cell_num), workbook)
                       for cell_num in range(0, len(TransactionRow.HEADER))]

                t_row = TransactionRow(row, row_num+1, worksheet.name)
                try:
                    t_row.parse()
                except TransactionParserError as e:
                    t_row.failure = e

                if t_row.failure:
                    log.error("%s", t_row)
                    log.error("%s", t_row.failure)

                self.t_rows.append(t_row)

        workbook.release_resources()
        del workbook

    @staticmethod
    def convert_cell(cell, workbook):
        if cell.ctype == xlrd.XL_CELL_DATE:
            value = xlrd.xldate.xldate_as_datetime(cell.value, workbook.datemode). \
                         strftime('%Y-%m-%d %H:%M:%S')
        elif cell.ctype == xlrd.XL_CELL_NUMBER:
            # repr is required to ensure no precision is lost
            value = repr(cell.value)
        else:
            value = str(cell.value)

        return value

    def import_csv(self, import_file):
        log.info("==IMPORT TRANSACTION RECORDS FROM CSV FILE: %s ==", import_file.name)

        if sys.version_info[0] < 3:
            # Special handling required for utf-8 encoded csv files
            reader = csv.reader(self.utf_8_encoder(import_file))
        else:
            reader = csv.reader(import_file)

        next(reader, None) # skip headers
        for row in reader:
            t_row = TransactionRow(row[:len(TransactionRow.HEADER)], reader.line_num)
            try:
                t_row.parse()
            except TransactionParserError as e:
                t_row.failure = e

            if t_row.failure:
                log.error("%s", t_row)
                log.error("%s", t_row.failure)

            self.t_rows.append(t_row)

    @staticmethod
    def utf_8_encoder(unicode_csv_data):
        for line in unicode_csv_data:
            yield line.encode('utf-8')

    def failures(self):
        return bool([t_row for t_row in self.t_rows if t_row.failure is not None])

    def get_records(self):
        transaction_records = [t_row.t_record for t_row in self.t_rows if t_row.t_record]

        transaction_records.sort()
        for t_record in transaction_records:
            t_record.set_tid()

        for t_row in self.t_rows:
            log.debug("%s", t_row)

        log.info("Total transaction records=%s", len(transaction_records))
        return transaction_records

class TransactionRow(object):
    HEADER = ['Type',
              'Buy Quantity', 'Buy Asset', 'Buy Value',
              'Sell Quantity', 'Sell Asset', 'Sell Value',
              'Fee Quantity', 'Fee Asset', 'Fee Value',
              'Wallet', 'Timestamp']

    BUY_TYPES = (TransactionRecord.TYPE_DEPOSIT,
                 TransactionRecord.TYPE_MINING,
                 TransactionRecord.TYPE_INCOME,
                 TransactionRecord.TYPE_GIFT_RECEIVED)
    SELL_TYPES = (TransactionRecord.TYPE_WITHDRAWAL,
                  TransactionRecord.TYPE_SPEND,
                  TransactionRecord.TYPE_GIFT_SENT,
                  TransactionRecord.TYPE_CHARITY_SENT)

    TRANSFER_TYPES = (TransactionRecord.TYPE_DEPOSIT,
                      TransactionRecord.TYPE_WITHDRAWAL)

    cnt = 0

    def __init__(self, row, row_num, worksheet_name=None):
        self.row = row
        self.row_num = row_num
        self.worksheet_name = worksheet_name
        self.t_record = None
        self.failure = None

    def parse(self):
        if all(not self.row[i] for i in range(len(self.HEADER))):
            # Skip empty rows
            return

        buy = sell = fee = None
        buy_asset = sell_asset = fee_asset = None

        t_type = self.row[0]
        if t_type in self.BUY_TYPES:
            buy_quantity, buy_asset, buy_value = self.validate_buy(self.row)
            self.validate_no_sell(self.row)
            fee_quantity, fee_asset, fee_value = self.validate_fee(self.row)
        elif t_type in self.SELL_TYPES:
            self.validate_no_buy(self.row)
            sell_quantity, sell_asset, sell_value = self.validate_sell(self.row)
            fee_quantity, fee_asset, fee_value = self.validate_fee(self.row)
        elif t_type == TransactionRecord.TYPE_TRADE:
            buy_quantity, buy_asset, buy_value = self.validate_buy(self.row)
            sell_quantity, sell_asset, sell_value = self.validate_sell(self.row)
            fee_quantity, fee_asset, fee_value = self.validate_fee(self.row)
        else:
            raise UnexpectedTransactionTypeError(0, self.HEADER[0], t_type)

        if buy_asset:
            buy = Buy(t_type, buy_quantity, buy_asset, buy_value)
        if sell_asset:
            sell = Sell(t_type, sell_quantity, sell_asset, sell_value)
        if fee_asset:
            # Fees are added as a separate spend transaction
            fee = Sell(TransactionRecord.TYPE_SPEND, fee_quantity, fee_asset, fee_value)
            if t_type in self.TRANSFER_TYPES:
                # A fee spend is normally a disposal unless it's part of a transfer
                fee.disposal = False

        self.t_record = TransactionRecord(t_type, buy, sell, fee, self.row[10],
                                          self.parse_timestamp(self.row[11]))

    @staticmethod
    def parse_timestamp(timestamp_str):
        try:
            timestamp = dateutil.parser.parse(timestamp_str, tzinfos=config.TZ_INFOS)
        except ValueError:
            raise TimestampParserError(11, TransactionRow.HEADER[11], timestamp_str)

        if timestamp.tzinfo is None:
            # Default to UTC if no timezone is specified
            timestamp = timestamp.replace(tzinfo=config.TZ_UTC)

        return timestamp

    @staticmethod
    def validate_buy(row):
        if row[1]:
            try:
                buy_quantity = Decimal(TransactionRow.strip_non_digits(row[1]))
            except InvalidOperation:
                raise DataValueError(1, TransactionRow.HEADER[1], row[1])

            if buy_quantity < 0:
                raise DataValueError(1, TransactionRow.HEADER[1], buy_quantity)
        else:
            raise MissingDataError(1, TransactionRow.HEADER[1])

        if row[2]:
            buy_asset = row[2]
        else:
            raise MissingDataError(2, TransactionRow.HEADER[2])

        if row[3]:
            try:
                buy_value = Decimal(TransactionRow.strip_non_digits(row[3]))
            except InvalidOperation:
                raise DataValueError(3, TransactionRow.HEADER[3], row[3])

            if buy_value < 0:
                raise DataValueError(3, TransactionRow.HEADER[3], buy_value)

            if buy_asset == config.CCY and buy_value != buy_quantity:
                raise DataValueError(3, TransactionRow.HEADER[3], buy_value)
        else:
            buy_value = None

        return buy_quantity, buy_asset, buy_value

    @staticmethod
    def validate_no_buy(row):
        if row[1]:
            raise UnexpectedDataError(1, TransactionRow.HEADER[1], row[1])

        if row[2]:
            raise UnexpectedDataError(2, TransactionRow.HEADER[2], row[2])

        if row[3]:
            raise UnexpectedDataError(3, TransactionRow.HEADER[3], row[3])

    @staticmethod
    def validate_sell(row):
        if row[4]:
            try:
                sell_quantity = Decimal(TransactionRow.strip_non_digits(row[4]))
            except InvalidOperation:
                raise DataValueError(4, TransactionRow.HEADER[4], row[4])

            if sell_quantity < 0:
                raise DataValueError(4, TransactionRow.HEADER[4], sell_quantity)
        else:
            raise MissingDataError(4, TransactionRow.HEADER[4])

        if row[5]:
            sell_asset = row[5]
        else:
            raise MissingDataError(5, TransactionRow.HEADER[5])

        if row[6]:
            try:
                sell_value = Decimal(TransactionRow.strip_non_digits(row[6]))
            except InvalidOperation:
                raise DataValueError(6, TransactionRow.HEADER[6], row[6])

            if sell_value < 0:
                raise DataValueError(6, TransactionRow.HEADER[6], sell_value)

            if sell_asset == config.CCY and sell_value != sell_quantity:
                raise DataValueError(6, TransactionRow.HEADER[6], sell_value)
        else:
            sell_value = None

        return sell_quantity, sell_asset, sell_value

    @staticmethod
    def validate_no_sell(row):
        if row[4]:
            raise UnexpectedDataError(4, TransactionRow.HEADER[4], row[4])

        if row[5]:
            raise UnexpectedDataError(5, TransactionRow.HEADER[5], row[5])

        if row[6]:
            raise UnexpectedDataError(6, TransactionRow.HEADER[6], row[6])

    @staticmethod
    def validate_fee(row):
        if row[7]:
            try:
                fee_quantity = Decimal(TransactionRow.strip_non_digits(row[7]))
            except InvalidOperation:
                raise DataValueError(7, TransactionRow.HEADER[7], row[7])

            if fee_quantity < 0:
                raise DataValueError(7, TransactionRow.HEADER[7], fee_quantity)
        else:
            fee_quantity = None

        if row[8]:
            fee_asset = row[8]
        else:
            fee_asset = None

        if row[9]:
            try:
                fee_value = Decimal(TransactionRow.strip_non_digits(row[9]))
            except InvalidOperation:
                raise DataValueError(9, TransactionRow.HEADER[9], row[9])

            if fee_value < 0:
                raise DataValueError(9, TransactionRow.HEADER[9], fee_value)

            if fee_asset == config.CCY and fee_value != fee_quantity:
                raise DataValueError(9, TransactionRow.HEADER[9], fee_value)
        else:
            fee_value = None

        if fee_quantity is not None and not fee_asset:
            raise MissingDataError(8, TransactionRow.HEADER[8])
        elif fee_quantity is None and fee_asset:
            raise MissingDataError(7, TransactionRow.HEADER[7])

        return fee_quantity, fee_asset, fee_value

    @staticmethod
    def strip_non_digits(string):
        return string.strip('£€$').replace(',', '')

    def __str__(self):
        if self.t_record and self.t_record.tid:
            tid_str = " TID:" + str(self.t_record.tid[0])
        else:
            tid_str = ""

        if self.worksheet_name:
            return "'" + self.worksheet_name + \
                   "' Row[" + str(self.row_num) + "]: " + \
                   '[' + '\'{0}\''.format('\', \''.join(self.row)) + ']' + \
                   tid_str

        return "Row[" + str(self.row_num) + "]: " + \
               '[' + '\'{0}\''.format('\', \''.join(self.row)) + ']' + \
               tid_str



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

import logging
import os
import platform
import pkg_resources

import yaml
import dateutil.tz

from .version import __version__

log = logging.getLogger()

class Config(object):
    TZ_INFOS = {'BST': dateutil.tz.gettz('Europe/London'),
                'GMT': dateutil.tz.gettz('Europe/London')}
    TZ_LOCAL = dateutil.tz.gettz('Europe/London')
    TZ_UTC = dateutil.tz.UTC
    CCY = 'GBP'

    BITTYTAX_PATH = os.path.expanduser('~/.bittytax')
    BITTYTAX_CONFIG = 'bittytax.conf'
    CACHE_DIR = os.path.join(BITTYTAX_PATH, 'cache')

    FIAT_LIST = ['GBP', 'EUR', 'USD']
    CRYPTO_LIST = ['BTC', 'ETH', 'XRP', 'LTC', 'BCH', 'USDT']

    FORMAT_CSV = 'CSV'
    FORMAT_EXCEL = 'EXCEL'
    FORMAT_RECAP = 'RECAP'

    TRADE_ASSET_TYPE_BUY = 0
    TRADE_ASSET_TYPE_SELL = 1
    TRADE_ASSET_TYPE_PRIORITY = 2

    DATA_SOURCE_FIAT = ["ExchangeRatesAPI", "RatesAPI"]
    DATA_SOURCE_CRYPTO = ["CryptoCompare", "CoinGecko"]

    DEFAULT_CONFIG = {
        'fiat_list': FIAT_LIST,
        'crypto_list': CRYPTO_LIST,
        'trade_asset_type': TRADE_ASSET_TYPE_PRIORITY,
        'show_empty_wallets': False,
        'transfers_include': True,
        'data_source_select': {},
        'data_source_fiat': DATA_SOURCE_FIAT,
        'data_source_crypto': DATA_SOURCE_CRYPTO,
    }

    def __init__(self):
        self.args = None

        if not os.path.exists(Config.BITTYTAX_PATH):
            os.mkdir(Config.BITTYTAX_PATH)

        if not os.path.exists(os.path.join(Config.BITTYTAX_PATH, Config.BITTYTAX_CONFIG)):
            default_conf = pkg_resources.resource_string(__name__,
                                                         'config/' + Config.BITTYTAX_CONFIG)
            with open(os.path.join(Config.BITTYTAX_PATH,
                                   Config.BITTYTAX_CONFIG), 'wb') as config_file:
                config_file.write(default_conf)

        try:
            with open(os.path.join(Config.BITTYTAX_PATH,
                                   Config.BITTYTAX_CONFIG), "rb") as config_file:
                self.config = yaml.safe_load(config_file)
        except:
            self.log.warning("Config file cannot be loaded: %s",
                             os.path.join(Config.BITTYTAX_PATH, Config.BITTYTAX_CONFIG))
            self.config = {}

        for name, default in self.DEFAULT_CONFIG.items():
            if name not in self.config:
                self.config[name] = default

        self.asset_priority = self.config['fiat_list'] + self.config['crypto_list']

    def __getattr__(self, name):
        try:
            return self.config[name]
        except KeyError:
            return getattr(self.args, name)

    def output_config(self, progname):
        log.debug("BITTYTAX: %s v%s", progname, __version__)
        log.debug("PYTHON: v%s SYSTEM: %s RELEASE: %s",
                  platform.python_version(), platform.system(), platform.release())

        for name in sorted(self.DEFAULT_CONFIG):
            log.debug("CONFIG: %s = %s", name, self.config[name])

    def sym(self):
        if self.CCY == 'GBP':
            return u'\xA3' # £
        else:
            raise ValueError("Currency not supported")

config = Config()



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

import logging
import datetime

import dateutil.parser
import dateutil.tz

from ..config import config

TERM_WIDTH = 69

log = logging.getLogger()

class DataParser(object):
    TYPE_WALLET = "Wallets"
    TYPE_EXCHANGE = "Exchanges"
    TYPE_EXPLORER = "Explorers"
    TYPE_SHARES = "Stocks & Shares"

    LIST_ORDER = (TYPE_WALLET, TYPE_EXCHANGE, TYPE_EXPLORER, TYPE_SHARES)

    parsers = []

    def __init__(self, p_type, name, header, delimiter=',',
                 worksheet_name=None, row_handler=None, all_handler=None):
        self.p_type = p_type
        self.name = name
        self.header = header
        self.worksheet_name = worksheet_name if worksheet_name else name
        self.delimiter = delimiter
        self.row_handler = row_handler
        self.all_handler = all_handler
        self.args = []
        self.in_header = None
        self.in_header_row_num = None

        self.parsers.append(self)

    def __eq__(self, other):
        return self.name == other.name

    def __ne__(self, other):
        return not self == other

    def __lt__(self, other):
        return self.name < other.name

    def format_header(self):
        header = []
        for col in self.header:
            if callable(col) or col is None:
                header.append('_')
            else:
                header.append(col)

        header_str = '\'' + str(self.delimiter).join(header) + '\''

        return header_str[:TERM_WIDTH] + '...' if len(header_str) > TERM_WIDTH else header_str

    @classmethod
    def parse_timestamp(cls, timestamp_str, tzinfos=None, tz=None, dayfirst=False):
        if isinstance(timestamp_str, int):
            timestamp = datetime.datetime.utcfromtimestamp(timestamp_str)
        else:
            timestamp = dateutil.parser.parse(timestamp_str, tzinfos=tzinfos, dayfirst=dayfirst)

        if tz:
            timestamp = timestamp.replace(tzinfo=dateutil.tz.gettz(tz))
            timestamp = timestamp.astimezone(config.TZ_UTC)
        elif timestamp.tzinfo is None:
            #default to UTC if no timezone is specified
            timestamp = timestamp.replace(tzinfo=config.TZ_UTC)
        else:
            timestamp = timestamp.astimezone(config.TZ_UTC)

        return timestamp

    @classmethod
    def match_header(cls, row, row_num):
        log.debug("Row[%s] TRY: %s", row_num+1, cls.format_row(row))
        parsers_reduced = [p for p in cls.parsers if len(p.header) == len(row)]
        for parser in parsers_reduced:
            match = False
            for i, row_field in enumerate(row):
                if callable(parser.header[i]):
                    match = parser.header[i](row_field)
                    parser.args.append(match)
                elif parser.header[i] is not None:
                    match = row_field == parser.header[i]

                if not match:
                    break

            if match:
                log.debug("Row[%s] MATCHED: %s '%s'",
                          row_num+1, cls.format_row(parser.header), parser.name)
                parser.in_header = row
                parser.in_header_row_num = row_num + 1
                return parser
            else:
                log.debug("Row[%s] NO MATCH: %s '%s'",
                          row_num+1, cls.format_row(parser.header), parser.name)

        raise KeyError

    @classmethod
    def format_parsers(cls):
        txt = ""
        for p_type in cls.LIST_ORDER:
            txt += ' ' * 2 + p_type + ':\n'
            prev_name = None
            for parser in sorted([parser for parser in cls.parsers if parser.p_type == p_type]):
                if parser.name != prev_name:
                    txt += ' ' * 4 + parser.name + '\n'
                txt += ' ' * 6 + parser.format_header() + '\n'

                prev_name = parser.name

        return txt

    @staticmethod
    def format_row(row):
        row_out = []
        for col in row:
            if callable(col):
                row_out.append('<lambda>')
            elif col is None:
                row_out.append('*')
            else:
                row_out.append('\'{}\''.format(col))

        return '[' + ', '.join(row_out) + ']'



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

import logging

from decimal import Decimal

from .config import config

log = logging.getLogger()

class AuditRecords(object):
    def __init__(self, transaction_records):
        self.wallets = {}
        self.totals = {}

        log.info("==AUDIT TRANSACTION RECORDS==")
        for tr in transaction_records:
            log.debug(tr)
            if tr.buy:
                self._add_tokens(tr.wallet, tr.buy.asset, tr.buy.quantity)

            if tr.sell:
                self._subtract_tokens(tr.wallet, tr.sell.asset, tr.sell.quantity)

            if tr.fee:
                self._subtract_tokens(tr.wallet, tr.fee.asset, tr.fee.quantity)

        log.debug("==TOTAL BALANCES==")
        for asset in sorted(self.totals):
            log.debug("%s=%s",
                      asset,
                      '{:0,f}'.format(self.totals[asset].normalize()))

    def _add_tokens(self, wallet, asset, quantity):
        if wallet not in self.wallets:
            self.wallets[wallet] = {}

        if asset not in self.wallets[wallet]:
            self.wallets[wallet][asset] = Decimal(0)

        self.wallets[wallet][asset] += quantity

        if asset not in self.totals:
            self.totals[asset] = Decimal(0)

        self.totals[asset] += quantity

        log.debug("%s:%s=%s (+%s)",
                  wallet,
                  asset,
                  '{:0,f}'.format(self.wallets[wallet][asset].normalize()),
                  '{:0,f}'.format(quantity.normalize()))

    def _subtract_tokens(self, wallet, asset, quantity):
        if wallet not in self.wallets:
            self.wallets[wallet] = {}

        if asset not in self.wallets[wallet]:
            self.wallets[wallet][asset] = Decimal(0)

        self.wallets[wallet][asset] -= quantity

        if asset not in self.totals:
            self.totals[asset] = Decimal(0)

        self.totals[asset] -= quantity

        log.debug("%s:%s=%s (-%s)",
                  wallet,
                  asset,
                  '{:0,f}'.format(self.wallets[wallet][asset].normalize()),
                  '{:0,f}'.format(quantity.normalize()))

        if self.wallets[wallet][asset] < 0 and asset not in config.fiat_list:
            log.warning("Balance at %s:%s is negative %s",
                        wallet,
                        asset,
                        '{:0,f}'.format(self.wallets[wallet][asset].normalize()))



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

import logging
import csv
import sys
import os

from ..config import config
from .out_record import TransactionOutRecord

log = logging.getLogger()

class OutputBase(object):
    DEFAULT_FILENAME = 'BittyTax_Records'
    EXCEL_PRECISION = 15
    BITTYTAX_OUT_HEADER = ['Type',
                           'Buy Quantity', 'Buy Asset', 'Buy Value',
                           'Sell Quantity', 'Sell Asset', 'Sell Value',
                           'Fee Quantity', 'Fee Asset', 'Fee Value',
                           'Wallet', 'Timestamp']

    RECAP_OUT_HEADER = ['Type', 'Date',
                        'InOrBuyAmount', 'InOrBuyCurrency',
                        'OutOrSellAmount', 'OutOrSellCurrency',
                        'FeeAmount', 'FeeCurrency']

    def __init__(self, datafiles):
        self.data_files = datafiles

    def out_header(self):
        if config.args.format == config.FORMAT_RECAP:
            return self.RECAP_OUT_HEADER

        return self.BITTYTAX_OUT_HEADER

    def in_header(self, in_header):
        if config.args.format == config.FORMAT_RECAP:
            return [name if name not in self.out_header()
                    else name + '_' for name in in_header]

        return in_header

    @staticmethod
    def get_output_filename(extension_type):
        if config.args.output_filename:
            filepath, file_extension = os.path.splitext(config.args.output_filename)
            if file_extension != extension_type:
                filepath = filepath + '.' + extension_type
        else:
            filepath = OutputBase.DEFAULT_FILENAME + '.' + extension_type

        if not os.path.exists(filepath):
            return filepath

        filepath, file_extension = os.path.splitext(filepath)
        i = 2
        new_fname = '{}-{}{}'.format(filepath, i, file_extension)
        while os.path.exists(new_fname):
            i += 1
            new_fname = '{}-{}{}'.format(filepath, i, file_extension)

        return new_fname

class OutputCsv(OutputBase):
    FILE_EXTENSION = 'csv'
    RECAP_TYPE_MAPPING = {TransactionOutRecord.TYPE_DEPOSIT: 'Deposit',
                          TransactionOutRecord.TYPE_MINING: 'Mining',
                          TransactionOutRecord.TYPE_INCOME: 'Income',
                          TransactionOutRecord.TYPE_GIFT_RECEIVED: 'Gift',
                          TransactionOutRecord.TYPE_WITHDRAWAL: 'Withdrawal',
                          TransactionOutRecord.TYPE_SPEND: 'Purchase',
                          TransactionOutRecord.TYPE_GIFT_SENT: 'Gift',
                          TransactionOutRecord.TYPE_CHARITY_SENT: 'Donation',
                          TransactionOutRecord.TYPE_TRADE: 'Trade'}
    def write_csv(self):
        if config.args.output_filename:
            filename = self.get_output_filename(self.FILE_EXTENSION)

            if sys.version_info[0] >= 3:
                with open(filename, 'w', newline='', encoding='utf-8') as csv_file:
                    writer = csv.writer(csv_file, lineterminator='\n')
                    self.write_rows(writer)
            else:
                with open(filename, 'wb') as csv_file:
                    writer = csv.writer(csv_file, lineterminator='\n')
                    self.write_rows(writer)

            log.info("Output CSV file created: %s", filename)
        else:
            if sys.version_info[0] >= 3:
                sys.stdout.reconfigure(encoding='utf-8')

            writer = csv.writer(sys.stdout, lineterminator='\n')
            self.write_rows(writer)

    def write_rows(self, writer):
        data_rows = []
        for data_file in self.data_files:
            data_rows.extend(data_file.data_rows)

        if config.args.sort:
            data_rows = sorted(data_rows, key=lambda dr: dr.timestamp, reverse=False)

        if not config.args.noheader:
            if config.args.append:
                writer.writerow(self.out_header() +
                                self.in_header(self.data_files[0].parser.in_header))
            else:
                writer.writerow(self.out_header())

        for data_row in data_rows:
            if config.args.append:
                if data_row.t_record:
                    writer.writerow(self._to_csv(data_row.t_record) + data_row.in_row)
                else:
                    writer.writerow([None] * len(self.out_header()) + data_row.in_row)
            else:
                if data_row.t_record:
                    writer.writerow(self._to_csv(data_row.t_record))

    def _to_csv(self, t_record):
        if config.args.format == config.FORMAT_RECAP:
            return self._to_recap_csv(t_record)

        return self._to_bittytax_csv(t_record)

    @staticmethod
    def _to_bittytax_csv(tr):
        if tr.buy_quantity is not None and \
                len(tr.buy_quantity.normalize().as_tuple().digits) > OutputBase.EXCEL_PRECISION:
            log.warning("%d-digit precision exceeded for Buy Quantity: %s",
                        OutputBase.EXCEL_PRECISION, tr.format_quantity(tr.buy_quantity))

        if tr.sell_quantity is not None and \
                len(tr.sell_quantity.normalize().as_tuple().digits) > OutputBase.EXCEL_PRECISION:
            log.warning("%d-digit precision exceeded for Sell Quantity: %s",
                        OutputBase.EXCEL_PRECISION, tr.format_quantity(tr.sell_quantity))

        if tr.fee_quantity is not None and \
                len(tr.fee_quantity.normalize().as_tuple().digits) > OutputBase.EXCEL_PRECISION:
            log.warning("%d-digit precision exceeded for Fee Quantity: %s",
                        OutputBase.EXCEL_PRECISION, tr.format_quantity(tr.fee_quantity))

        return [tr.t_type,
                '{0:f}'.format(tr.buy_quantity.normalize()) if tr.buy_quantity is not None \
                                                            else None,
                tr.buy_asset,
                '{0:f}'.format(tr.buy_value.normalize()) if tr.buy_value is not None \
                                                         else None,
                '{0:f}'.format(tr.sell_quantity.normalize()) if tr.sell_quantity is not None \
                                                             else None,
                tr.sell_asset,
                '{0:f}'.format(tr.sell_value.normalize()) if tr.sell_value is not None \
                                                          else None,
                '{0:f}'.format(tr.fee_quantity.normalize()) if tr.fee_quantity is not None \
                                                            else None,
                tr.fee_asset,
                '{0:f}'.format(tr.fee_value.normalize()) if tr.fee_value is not None \
                                                         else None,
                tr.wallet,
                tr.timestamp.strftime('%Y-%m-%dT%H:%M:%S %Z')]

    @staticmethod
    def _to_recap_csv(tr):
        return [OutputCsv.RECAP_TYPE_MAPPING[tr.t_type],
                tr.timestamp.strftime('%Y-%m-%d %H:%M:%S'),
                '{0:f}'.format(tr.buy_quantity.normalize()) if tr.buy_quantity is not None \
                                                            else None,
                tr.buy_asset,
                '{0:f}'.format(tr.sell_quantity.normalize()) if tr.sell_quantity is not None \
                                                             else None,
                tr.sell_asset,
                '{0:f}'.format(tr.fee_quantity.normalize()) if tr.fee_quantity is not None \
                                                            else None,
                tr.fee_asset]



import os
import re
import io

from setuptools import setup

BITTYTAX_PATH = os.path.expanduser('~/.bittytax')
VERSION_FILE = 'bittytax/version.py'
GITHUB_REPO = 'https://github.com/BittyTax/BittyTax'

def get_version():
    line = open(VERSION_FILE, 'rt').read()
    version = re.search(r'^__version__ = [\'"]([^\'"]*)[\'"]', line, re.MULTILINE)
    if version:
        return version.group(1)
    else:
        raise RuntimeError('Unable to find version string in ' + VERSION_FILE)

def get_long_description():
    with io.open('README.md', encoding='utf8') as ld:
        return ld.read()

setup(
    name='BittyTax',
    version=get_version(),
    description='Cryptoasset accounting, auditing and UK tax calculations (Capital Gains/Income '
                'Tax)',
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    url=GITHUB_REPO,
    download_url=GITHUB_REPO + '/archive/v{}.zip'.format(get_version()),
    author='Scott Green/Nano Nano Ltd',
    author_email='scott@bitty.tax',
    license='AGPLv3',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'Topic :: Office/Business :: Financial :: Accounting',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Natural Language :: English',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    keywords='bittytax cryptoasset cryptocurrency crypto tax',
    packages=['bittytax', 'bittytax.conv', 'bittytax.price'],
    package_data={'bittytax': ['templates/*.html']},
    install_requires=[
        'python-dateutil',
        'requests',
        'pyyaml',
        'xlrd',
        'xlsxwriter',
        'jinja2',
        'xhtml2pdf',
    ],
    entry_points={
        'console_scripts': [
            'bittytax = bittytax.bittytax:main',
            'bittytax_conv = bittytax.conv.bittytax_conv:main',
            'bittytax_price = bittytax.price.bittytax_price:main',
        ],
    },
    include_package_data=True,
    zip_safe=False,
    )



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

import logging
import os

from ..version import __version__
from ..config import config
from .datasource import ExchangeRatesAPI, RatesAPI, CoinDesk, CryptoCompare, CoinGecko, CoinPaprika

log = logging.getLogger()

class PriceData(object):
    def __init__(self):
        self.data_sources = {}

        if not os.path.exists(config.CACHE_DIR):
            os.mkdir(config.CACHE_DIR)

        data_sources_required = set(config.data_source_fiat +
                                    config.data_source_crypto) | \
                                {x for v in config.data_source_select.values() for x in v}

        for data_source in data_sources_required:
            try:
                self.data_sources[data_source] = globals()[data_source]()
            except KeyError:
                raise ValueError("Data source: %s not recognised" % [data_source])

    @staticmethod
    def data_source_priority(asset):
        if asset in config.data_source_select:
            return config.data_source_select[asset]
        elif asset in config.fiat_list:
            return config.data_source_fiat
        else:
            return config.data_source_crypto

    def get_latest_ds(self, data_source, asset, quote):
        if data_source in self.data_sources:
            if asset in self.data_sources[data_source].assets:
                return self.data_sources[data_source].get_latest(asset, quote), \
                       self.data_sources[data_source].assets[asset]

            return None, None
        else:
            raise ValueError("Data source: %s not recognised" % [data_source])

    def get_historical_ds(self, data_source, asset, quote, timestamp):
        if data_source in self.data_sources:
            if asset in self.data_sources[data_source].assets:
                date = timestamp.strftime('%Y-%m-%d')
                pair = asset + '/' + quote

                if not config.args.nocache:
                    # check cache first
                    if pair in self.data_sources[data_source].prices and \
                       date in self.data_sources[data_source].prices[pair]:
                        return self.data_sources[data_source].prices[pair][date]['price'], \
                               self.data_sources[data_source].assets[asset], \
                               self.data_sources[data_source].prices[pair][date]['url']

                self.data_sources[data_source].get_historical(asset, quote, timestamp)
                if pair in self.data_sources[data_source].prices and \
                   date in self.data_sources[data_source].prices[pair]:
                    return self.data_sources[data_source].prices[pair][date]['price'], \
                           self.data_sources[data_source].assets[asset], \
                           self.data_sources[data_source].prices[pair][date]['url']
                else:
                    return None, self.data_sources[data_source].assets[asset], None
            else:
                return None, None, None
        else:
            raise ValueError("Data source: %s not recognised" % [data_source])

    def get_latest(self, asset, quote):
        price = name = data_source = None
        for data_source in self.data_source_priority(asset):
            price, name = self.get_latest_ds(data_source, asset, quote)
            if price is not None:
                log.debug("Price (latest), 1 %s=%s %s via %s (%s)",
                          asset,
                          '{:0,f}'.format(price.normalize()),
                          quote,
                          data_source,
                          name)
                break

        return price, name, data_source

    def get_historical(self, asset, quote, timestamp):
        price = name = data_source = None
        for data_source in self.data_source_priority(asset):
            price, name, url = self.get_historical_ds(data_source, asset, quote, timestamp)
            if price is not None:
                log.debug("Price on %s, 1 %s=%s %s via %s (%s)",
                          timestamp.strftime('%Y-%m-%d'),
                          asset,
                          '{:0,f}'.format(price.normalize()),
                          quote,
                          data_source,
                          name)
                break

        return price, name, data_source, url



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

import logging
from decimal import Decimal
from datetime import datetime

from ..version import __version__
from ..config import config
from ..tax import which_tax_year
from .pricedata import PriceData

log = logging.getLogger()

class ValueAsset(object):
    def __init__(self):
        self.price_data = PriceData()
        self.price_report = {}

    def get_value(self, asset, timestamp, quantity):
        if asset == config.CCY:
            return quantity, True

        asset_price_ccy, _, _ = self.get_historical_price(asset, timestamp)
        if asset_price_ccy is not None:
            value = asset_price_ccy * quantity
            log.debug("Price on %s, 1 %s=%s%s %s, %s %s=%s%s %s",
                      timestamp.strftime('%Y-%m-%d'),
                      asset,
                      config.sym(), '{:0,.2f}'.format(asset_price_ccy), config.CCY,
                      '{:0,f}'.format(quantity),
                      asset,
                      config.sym(), '{:0,.2f}'.format(value), config.CCY)
            return value, False
        else:
            log.debug("Price at %s for %s is not available",
                      timestamp.strftime('%Y-%m-%d'), asset)
            return Decimal(0), False

    def get_current_value(self, asset, quantity):
        asset_price_ccy, name, data_source = self.get_latest_price(asset)
        if asset_price_ccy is not None:
            return asset_price_ccy * quantity, name, data_source

        return None, None, None

    def get_historical_price(self, asset, timestamp):
        asset_price_ccy = None

        if timestamp.date() >= datetime.now().date():
            raise Exception("Date is not in the past: " + timestamp.strftime('%Y-%m-%d'))

        if asset == "BTC" or asset in config.fiat_list:
            asset_price_ccy, name, data_source, url = self.price_data.get_historical(asset,
                                                                                     config.CCY,
                                                                                     timestamp)
            self.price_report_cache(asset, timestamp, name, data_source, url, asset_price_ccy)
        else:
            btc_price_ccy, name, data_source, url = self.price_data.get_historical("BTC",
                                                                                   config.CCY,
                                                                                   timestamp)
            self.price_report_cache("BTC", timestamp, name, data_source, url, btc_price_ccy)

            if btc_price_ccy is not None:
                asset_price_btc, name, data_source, url = self.price_data.get_historical(asset,
                                                                                         "BTC",
                                                                                         timestamp)
                if asset_price_btc is not None:
                    asset_price_ccy = btc_price_ccy * asset_price_btc

                self.price_report_cache(asset, timestamp, name, data_source, url,
                                        asset_price_ccy, asset_price_btc)

        return asset_price_ccy, name, data_source

    def get_latest_price(self, asset):
        asset_price_ccy = None

        if asset == "BTC" or asset in config.fiat_list:
            asset_price_ccy, name, data_source = self.price_data.get_latest(asset, config.CCY)
        else:
            btc_price_ccy, _, _ = self.price_data.get_latest("BTC", config.CCY)

            if btc_price_ccy is not None:
                asset_price_btc, name, data_source = self.price_data.get_latest(asset, "BTC")
                if asset_price_btc is not None:
                    asset_price_ccy = btc_price_ccy * asset_price_btc

        return asset_price_ccy, name, data_source

    def price_report_cache(self, asset, timestamp, name, data_source, url,
                           price_ccy, price_btc=None):
        tax_year = which_tax_year(timestamp)

        if tax_year not in self.price_report:
            self.price_report[tax_year] = {}

        if asset not in self.price_report[tax_year]:
            self.price_report[tax_year][asset] = {}

        date = timestamp.strftime('%Y-%m-%d')
        if date not in self.price_report[tax_year][asset]:
            self.price_report[tax_year][asset][date] = {'name': name,
                                                        'data_source': data_source,
                                                        'url': url,
                                                        'price_ccy': price_ccy,
                                                        'price_btc': price_btc}



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

import logging
import os
import atexit
import json
from decimal import Decimal
from datetime import datetime, timedelta

import dateutil.parser
import requests

from ..version import __version__
from ..config import config

CRYPTOCOMPARE_MAX_DAYS = 2000
COINPAPRIKA_MAX_DAYS = 5000

log = logging.getLogger()

class DataSourceBase(object):
    USER_AGENT = 'BittyTax/v{}'.format(__version__)
    TIME_OUT = 20

    def __init__(self):
        self.assets = {}
        self.prices = self.load_prices()

        for pair in sorted(self.prices):
            log.debug("PRICE DATA: %s (%s) loaded", self.name(), pair)

        atexit.register(self.dump_prices)

    def name(self):
        return self.__class__.__name__

    def get_json(self, url):
        log.debug(url)
        response = requests.get(url, headers={'User-Agent': self.USER_AGENT}, timeout=self.TIME_OUT)

        if response.status_code in [429, 502, 503, 504]:
            response.raise_for_status()

        return response.json()

    def update_prices(self, pair, prices, timestamp):
        if pair not in self.prices:
            self.prices[pair] = {}

        # We are not interested in today's latest price, only the days closing price, also need to
        #  filter any erroneous future dates returned
        prices = {k: v
                  for k, v in prices.items()
                  if dateutil.parser.parse(k).date() < datetime.now().date()}

        # We might not receive data for the date requested, if so set to None to prevent repeat
        #  lookups, assuming date is in the past
        date = timestamp.strftime('%Y-%m-%d')
        if date not in prices and timestamp.date() < datetime.now().date():
            prices[date] = {'price': None,
                            'url': None}

        self.prices[pair].update(prices)

    def load_prices(self):
        try:
            with open(os.path.join(config.CACHE_DIR, self.name() + '.json'), "r") as price_cache:
                json_prices = json.load(price_cache)
                return {pair: {date: {'price': self.str_to_decimal(price['price']),
                                      'url': price['url']}
                               for date, price in json_prices[pair].items()}
                        for pair in json_prices}
        except:
            log.warning("Price data for %s cannot be loaded", self.name())
            return {}

    def dump_prices(self):
        with open(os.path.join(config.CACHE_DIR, self.name() + '.json'), 'w') as price_cache:
            json_prices = {pair: {date: {'price': self.decimal_to_str(price['price']),
                                         'url': price['url']}
                                  for date, price in self.prices[pair].items()}
                           for pair in self.prices}
            json.dump(json_prices, price_cache, indent=4, sort_keys=True)

    @staticmethod
    def pair(asset, quote):
        return asset + '/' + quote

    @staticmethod
    def str_to_decimal(price):
        if price:
            return Decimal(price)

        return None

    @staticmethod
    def decimal_to_str(price):
        if price:
            return '{0:f}'.format(price)

        return None

    @staticmethod
    def epoch_time(timestamp):
        epoch = (timestamp - datetime(1970, 1, 1, tzinfo=config.TZ_UTC)).total_seconds()
        return int(epoch)

class ExchangeRatesAPI(DataSourceBase):
    def __init__(self):
        super(ExchangeRatesAPI, self).__init__()
        currencies = ['EUR', 'USD', 'JPY', 'BGN', 'CYP', 'CZK', 'DKK', 'EEK', 'GBP', 'HUF',
                      'LTL', 'LVL', 'MTL', 'PLN', 'ROL', 'RON', 'SEK', 'SIT', 'SKK', 'CHF',
                      'ISK', 'NOK', 'HRK', 'RUB', 'TRL', 'TRY', 'AUD', 'BRL', 'CAD', 'CNY',
                      'HKD', 'IDR', 'ILS', 'INR', 'KRW', 'MXN', 'MYR', 'NZD', 'PHP', 'SGD',
                      'THB', 'ZAR']
        self.assets = {c: 'Fiat' for c in currencies}

    def get_latest(self, asset, quote):
        url = 'https://api.exchangeratesapi.io/latest?base={}&symbols={}'.format(asset, quote)
        json_resp = self.get_json(url)
        return Decimal(repr(json_resp['rates'][quote])) if quote in json_resp['rates'] else None

    def get_historical(self, asset, quote, timestamp):
        url = 'https://api.exchangeratesapi.io/{}' \
              '?base={}&symbols={}'.format(timestamp.strftime('%Y-%m-%d'), asset, quote)
        json_resp = self.get_json(url)
        pair = self.pair(asset, quote)
        # Date returned in response might not be date requested due to weekends/holidays
        self.update_prices(pair,
                           {timestamp.strftime('%Y-%m-%d'): {
                               'price': Decimal(repr(json_resp['rates'][quote])) \
                                        if quote in json_resp['rates'] else None,
                               'url': url}},
                           timestamp)
        return self.prices[pair]

class RatesAPI(DataSourceBase):
    def __init__(self):
        super(RatesAPI, self).__init__()
        # https://github.com/MicroPyramid/ratesapi/blob/master/scripts/pusher.py
        currencies = ['EUR', 'USD', 'JPY', 'BGN', 'CYP', 'CZK', 'DKK', 'EEK', 'GBP', 'HUF',
                      'LTL', 'LVL', 'MTL', 'PLN', 'ROL', 'RON', 'SEK', 'SIT', 'SKK', 'CHF',
                      'ISK', 'NOK', 'HRK', 'RUB', 'TRL', 'TRY', 'AUD', 'BRL', 'CAD', 'CNY',
                      'HKD', 'IDR', 'ILS', 'INR', 'KRW', 'MXN', 'MYR', 'NZD', 'PHP', 'SGD',
                      'THB', 'ZAR']
        self.assets = {c: 'Fiat' for c in currencies}

    def get_latest(self, asset, quote):
        json_resp = self.get_json(
            'https://api.ratesapi.io/api/latest'
            '?base={}&symbols={}'.format(asset, quote)
        )
        return Decimal(repr(json_resp['rates'][quote])) if quote in json_resp['rates'] else None

    def get_historical(self, asset, quote, timestamp):
        url = 'https://api.ratesapi.io/api/{}' \
              '?base={}&symbols={}'.format(timestamp.strftime('%Y-%m-%d'), asset, quote)
        json_resp = self.get_json(url)
        pair = self.pair(asset, quote)
        # Date returned in response might not be date requested due to weekends/holidays
        self.update_prices(pair,
                           {timestamp.strftime('%Y-%m-%d'): {
                               'price': Decimal(repr(json_resp['rates'][quote])) \
                                        if quote in json_resp['rates'] else None,
                               'url': url}},
                           timestamp)
        return self.prices[pair]

class CoinDesk(DataSourceBase):
    def __init__(self):
        super(CoinDesk, self).__init__()
        self.assets = {"BTC": "Bitcoin"}

    def get_latest(self, _, quote):
        json_resp = self.get_json('https://api.coindesk.com/v1/bpi/currentprice.json')
        return Decimal(repr(json_resp['bpi'][quote]['rate_float'])) \
                if quote in json_resp['bpi'] else None

    def get_historical(self, asset, quote, timestamp):
        url = 'https://api.coindesk.com/v1/bpi/historical/close.json' \
              '?start={}&end={}&currency={}'.format(timestamp.strftime('%Y-%m-%d'), \
                                                    datetime.now().strftime('%Y-%m-%d'), quote)
        json_resp = self.get_json(url)
        pair = self.pair(asset, quote)
        self.update_prices(pair,
                           {k: {
                               'price': Decimal(repr(v)) if v else None,
                               'url': url} for k, v in json_resp['bpi'].items()}, timestamp)
        return self.prices[pair]

class CryptoCompare(DataSourceBase):
    def __init__(self):
        super(CryptoCompare, self).__init__()
        json_resp = self.get_json('https://min-api.cryptocompare.com/data/all/coinlist')
        self.assets = {c[1]['Symbol']: c[1]['CoinName'] for c in json_resp['Data'].items()}

    def get_latest(self, asset, quote):
        json_resp = self.get_json(
            'https://min-api.cryptocompare.com/data/price'
            '?extraParams={}&fsym={}&tsyms={}'.format(self.USER_AGENT, asset, quote))
        return Decimal(repr(json_resp[quote])) if quote in json_resp else None

    def get_historical(self, asset, quote, timestamp):
        url = 'https://min-api.cryptocompare.com/data/histoday' \
              '?aggregate=1&extraParams={}' \
              '&fsym={}&tsym={}&limit={}&tryConversion=false&toTs={}'.format(
                  self.USER_AGENT, asset, quote, CRYPTOCOMPARE_MAX_DAYS,
                  str(self.epoch_time(timestamp + timedelta(days=CRYPTOCOMPARE_MAX_DAYS))))

        json_resp = self.get_json(url)
        pair = self.pair(asset, quote)
        # Warning - CryptoCompare returns 0 as data for missing dates, convert these to None.
        self.update_prices(pair,
                           {datetime.fromtimestamp(d['time']).strftime('%Y-%m-%d'): {
                               'price': Decimal(repr(d['close'])) if d['close'] else None,
                               'url': url} for d in json_resp['Data']}, timestamp)
        return self.prices[pair]

class CoinGecko(DataSourceBase):
    def __init__(self):
        super(CoinGecko, self).__init__()
        json_resp = self.get_json('https://api.coingecko.com/api/v3/coins/list')
        self.assets = {c['symbol'].upper(): c['name'] for c in json_resp}
        self.ids = {c['symbol'].upper(): c['id'] for c in json_resp}

    def get_latest(self, asset, quote):
        json_resp = self.get_json(
            'https://api.coingecko.com/api/v3/coins/{}'
            '?localization=false&community_data=false&developer_data=false'.format(self.ids[asset]))
        return Decimal(repr(json_resp['market_data']['current_price'][quote.lower()])) \
                if quote.lower() in json_resp['market_data']['current_price'] else None

    def get_historical(self, asset, quote, timestamp):
        url = 'https://api.coingecko.com/api/v3/coins/{}/market_chart' \
              '?vs_currency={}&days=max'.format(self.ids[asset], quote)
        json_resp = self.get_json(url)
        pair = self.pair(asset, quote)
        self.update_prices(pair,
                           {datetime.utcfromtimestamp(p[0]/1000).strftime('%Y-%m-%d'): {
                               'price': Decimal(repr(p[1])) if p[1] else None,
                               'url': url} for p in json_resp['prices']},
                           timestamp)
        return self.prices[pair]

class CoinPaprika(DataSourceBase):
    def __init__(self):
        super(CoinPaprika, self).__init__()
        json_resp = self.get_json('https://api.coinpaprika.com/v1/coins')
        self.assets = {c['symbol']: c['name'] for c in json_resp}
        self.ids = {c['symbol']: c['id'] for c in json_resp}

    def get_latest(self, asset, quote):
        json_resp = self.get_json(
            'https://api.coinpaprika.com/v1/tickers/{}'
            '?quotes={}'.format(self.ids[asset], quote))
        return Decimal(repr(json_resp['quotes'][quote]['price'])) \
                if quote in json_resp['quotes'] else None

    def get_historical(self, asset, quote, timestamp):
        # Historic prices only available in USD or BTC
        if quote not in ('USD', 'BTC'):
            return None

        url = 'https://api.coinpaprika.com/v1/tickers/{}/historical' \
              '?start={}&limit={}&quote={}&interval=1d'.format(self.ids[asset],
                                                               timestamp.strftime('%Y-%m-%d'),
                                                               COINPAPRIKA_MAX_DAYS, quote)
        pair = self.pair(asset, quote)
        json_resp = self.get_json(url)
        pair = self.pair(asset, quote)
        self.update_prices(pair,
                           {dateutil.parser.parse(p['timestamp']).strftime('%Y-%m-%d'): {
                               'price': Decimal(repr(p['price'])) if p['price'] else None,
                               'url': url} for p in json_resp},
                           timestamp)
        return self.prices[pair]
